# 内存映射

内存映射是一种将文件或设备与内存区域建立映射关系的技术，它允许将文件或设备的内容直接映射到进程的虚拟内存空间中，以便程序可以像访问普通内存一样访问文件内容或设备资源。

即实现mmap&munmap系统调用

mmap_info结构体定义

```

struct mmap_info {
    int used;              // 标记是否已使用。1 表示该映射已经被使用，0 表示未使用。
    uint64 start;          // 映射的起始地址，表示内存映射的起始虚拟地址。
    uint64 len;            // 映射区域的长度，以字节为单位。
    int prot;              // 内存保护标志，例如 PROT_READ, PROT_WRITE, PROT_EXEC 等。
    int flags;             // 映射标志，用于指定内存映射类型（如 MAP_PRIVATE, MAP_SHARED 等）。
    int fd;                // 文件描述符，表示要映射的文件。如果是匿名映射，通常为 -1。
    off_t offset;          // 映射的文件偏移量，表示从文件的哪个位置开始映射。
};
```

我们的系统采用的是页式管理机制，主存空间按页分配。在`proc`进程结构体中增加`mmap_info`数组`mmap_pool`用于记录当前进程内存映射的信息。

```

struct proc {
  struct spinlock lock;
   ...
  struct mmap_info mmap_pool[MAXMMAP];
   ...
};

```

在实现`mmap`映射时，需要进行以下操作

1. 在当前进程的`mmap_pool`中找到可用的区域，用于保存`mmap`的信息，未找到则返回-1；
2. 根据文件描述符找到对应的文件，获取文件内部的`struct dirent`结构体指针`ep`；
3. 从给定的偏移开始，读取文件的内容，同时增长当前进程虚拟地址空间的大小，将文件的内容拷贝至新增的那段虚拟地址空间对应的物理内存中，根据`prot`参数设置内存空间的读、写与执行权限。
4. 将所使用的`mmap_info`的`used`字段置为1，同时将文件描述符`fd`对应文件的引用计数加1；

# 相关操作

##### 1.Map区域的初始化

```
for(int i = 0; i < MMAPNUM; ++i){
    p->mmap_pool[i].used = 0;
  }
```

在allocproc时调用

#####  2.find_unused_mmap_area

```
/**
 * @brief 查找未使用的内存映射区域
 * 
 * 该函数用于在当前进程的内存映射池中查找一个未使用的区域。
 * 
 * @param start 内存映射的起始地址
 * @param len 内存映射的长度
 * @param prot 内存映射的保护标志
 * @param flags 内存映射的标志
 * @param fd 文件描述符
 * @param offset 文件中的偏移量
 * @return int 返回找到的未使用区域的索引，如果没有找到则返回 -1
 */
static int find_unused_mmap_area(uint64 start,uint64 len,int prot,int flags,int fd,uint64 offset)
{
	// 初始化索引变量 i
	int i;
	// 获取当前进程的指针 p
	struct proc *p = myproc();
    // 遍历内存映射池中的每个区域
    for(i =0; i<MMAPNUM; i++)
    {
        // 如果当前区域未被使用，则返回该区域的索引
        if(p->mmap_pool[i].used == 0)
        { 
            return i;
        }
    }
    // 如果没有找到未使用的区域，则返回 -1
    return -1;
}
```

##### 3.find_mapped_area

```
/**
 * @brief 查找指定起始地址和长度的已映射内存区域的索引
 * 
 * 该函数遍历当前进程的内存映射池，查找与给定起始地址和长度匹配的已映射内存区域。
 * 
 * @param start 内存映射的起始地址
 * @param len 内存映射的长度
 * @return int 返回找到的已映射区域的索引，如果没有找到则返回 -1
 */
static int find_mapped_area(uint64 start,int len)
{
    // 获取当前进程的指针 p
    struct proc *p = myproc();
    // 遍历内存映射池中的每个区域
    for(int i =0; i<MMAPNUM; i++)
    {
        // 如果当前区域的起始地址和长度与给定的参数匹配，则返回该区域的索引
        if(p->mmap_pool[i].start == start && p->mmap_pool[i].len == len )
        {  
            return i;
        }
    }
    // 如果没有找到匹配的已映射区域，则返回 -1
    return -1;
}
```

##### 4.do_mmap

```
/**
 * @brief 执行内存映射操作
 * 
 * 该函数负责在当前进程的地址空间中创建一个新的内存映射。
 * 
 * @param start 内存映射的起始地址，如果为NULL，则使用当前进程的地址空间大小作为起始地址
 * @param len 内存映射的长度
 * @param prot 内存映射的保护标志，如读、写、执行等
 * @param flags 内存映射的标志，如共享、私有等
 * @param fd 文件描述符，指定要映射的文件
 * @param offset 文件中的偏移量，指定从文件的哪个位置开始映射
 * @return uint64 返回映射区域的起始地址，如果失败则返回 -1
 */
uint64 
do_mmap(uint64 start, uint64 len,int prot,int flags,int fd, off_t offset)
{   
    // 获取当前进程的指针 p
    struct proc *p = myproc();

    // 检查文件描述符和偏移量是否合法
    if(fd < 0 || fd >= NOFILE || offset <0)
        return -1;

    // 查找一个未使用的映射区域
    int index =  find_unused_mmap_area(start, len, prot, flags, fd, offset);

    if(-1 == index)
        return -1;// 如果没有找到可用的映射区域，返回错误

    // 设置映射区域的起始地址
    if(start == NULL)
    {   
        p->mmap_pool[index].start = p->sz;
    }
    else
    {
        p->mmap_pool[index].start = (uint64)start;
    }

    // 保存映射区域的相关信息
    p->mmap_pool[index].len = len;// 映射长度
    p->mmap_pool[index].prot = prot;// 保护标志
    p->mmap_pool[index].flags = flags;// 映射标志
    p->mmap_pool[index].fd = fd;// 文件描述符
    p->mmap_pool[index].offset = offset;// 偏移量

    struct dirent* ep = p->ofile[fd]->ep;// 获取文件目录项指针

    // 检查映射的范围是否超出文件大小
    if(offset + len > ep->file_size)
        return -1;
    uint tot_size = len;// 总大小

    // 计算需要映射的总页数
    int page_num = (tot_size + PGSIZE -1) / PGSIZE;

    // 设置页表权限
    int perm  = PTE_U;
    if(prot & PROT_READ) 
        perm  |= (PTE_R | PTE_A);
    if(prot & PROT_WRITE)
        perm  |= (PTE_W | PTE_D);
    if(prot & PROT_EXEC)
        perm  |= (PTE_X | PTE_A);
      
    uint64 va = p->mmap_pool[index].start, next_va;// 虚拟地址起始位置  
    uint64 off = offset;// 文件偏移量
    uint64 size;
    uint64 psize = p->sz;// 进程地址空间大小

    // 遍历每一页并进行内存映射
    for(int i = 0; i < page_num; i++)
    {
        size = tot_size <= PGSIZE ? tot_size : PGSIZE;// 计算当前页大小
        psize += PGSIZE;// 更新进程地址空间大小
        next_va = uvmalloc(p->pagetable, p->kpagetable, p->sz, psize);// 分配新的虚拟内存页
        if(next_va != psize){
            return -1;
        }
        p->sz = psize;// 更新进程地址空间大小
        uint64 pa = (uint64)walkaddr(p->pagetable, va);// 获取物理地址
        eread(ep, 0, pa, off, size);// 将文件内容读取到物理内存地址中
        
        if (i == page_num - 1) {
            memset((void*)(pa + size), 0, PGSIZE - size);// 如果是最后一页，填充未使用的部分
        }

        va = next_va;// 更新虚拟地址
        off+=size;// 更新偏移量
        tot_size -= size;// 更新剩余大小
        off += size;
    }
    

    p->mmap_pool[index].used = 1;// 标记映射区域为已使用
    filedup(p->ofile[p->mmap_pool[index].fd]);// 增加文件引用计数
    return p->mmap_pool[index].start;// 返回映射区域的起始地址
}
```

##### 5.do_munmap

```
uint64
do_munmap(uint64 start, uint64 len)
{
    // 找到映射区域的索引
    int index = find_mapped_area(start, len);
    if(index == -1)
        return -1;// 如果没有找到对应的映射区域，返回错误
    
    // 获取当前进程
    struct proc *p = myproc();
    // 获取对应的映射信息
    struct mmap_info *info = &(p->mmap_pool[index]);
    
    int tot_size = len;
    // 计算需要解除映射的页数
    int page_num = (tot_size + PGSIZE -1) / PGSIZE;
    int size;
    // 虚拟地址起始点
    uint64 va = start;
    // 获取文件系统中的目录项
    struct dirent *ep = p->ofile[info->fd]->ep;
    // 文件中的偏移量
    off_t off = info->offset;
    // 物理地址
    uint64 pa;

    for(int i = 0; i < page_num; i++)
    {
        size = tot_size <= PGSIZE ? tot_size : PGSIZE;// 计算每次操作的大小
        
        pte_t *pte = walk(p->pagetable, va, 0);// 获取页表项

        // 检查页表项是否有效
        if((*pte & PTE_V) == 0)
            return -1;
        if((*pte & PTE_U) == 0)
            return -1;
        // 如果页面被修改过，需要将其写回文件
        if(*pte & PTE_D){
            pa = PTE2PA(*pte);
                if(!pa){
                return -1;
            }
            ewrite(ep, 0, pa, off, size);// 将修改过的页面写回文件
        }

        // 从进程和内核的页表中解除映射
        vmunmap(p->pagetable, va, 1, 1);
        vmunmap(p->kpagetable, va, 1, 0);

        // 更新虚拟地址和文件偏移量
        va += PGSIZE;
        off += size;
        tot_size -= size;
    }

    // 标记映射信息为未使用，并关闭文件
    info->used = 0;
    fileclose(p->ofile[info->fd]);

    return 0;
}
```

