# 分页硬件

`RISC-V` 架构的指令在用户态和内核态下均提供了虚拟地址。物理内存（RAM）具有物理地址，用于唯一标记实际的内存空间。我们利用 `RISC-V` 的页表硬件来实现了虚拟地址到物理地址的转换。

- **架构**：我们的页表采用 **`Sv39`** 架构，它维护了一个**三级结构**的页表：

  - 第一级页表地址存储在 `satp` 寄存器，包含512个表项，每个表项指向下一级页表；

  - 第二级页表同样包含512个表项，每个表项指向第三级页表；

  - 第三级页表包含物理地址的物理页号 `PPN（Page Physical Number）`。

- ##### 地址转换过程

  虚拟地址的转换过程分为三个步骤：

  页目录索引（高 10 位）：虚拟地址的高 10 位作为页目录索引，从页目录中找到对应的页表。

  页表索引（中 10 位）：虚拟地址的中间 10 位作为页表索引，从页表中找到对应的物理页。

  页内偏移（低 12 位）：虚拟地址的低 12 位作为页内偏移，用于定位具体的物理地址。

- **页表缓存 （TLB）**：

由于 `CPU` 访问页表需要三次访存操作，影响访问速度。因此，硬件中存在 `TLB` ，类似于缓存，用于保存页表信息。 `CPU` 首先在 `TLB` 中查找页表元素，如果 `TLB` 命中，则停止访存；否则继续访存并更新 `TLB` 。

- **页表元素标志**：每个页表元素存储了标志位：

  - `PTE_V`：页表项是否有效；
  - `PTE_R`：页是否可读；

  - `PTE_W`：页是否可写；

  - `PTE_X`：页是否可执行；

  - `PTE_U`：用户态下是否可以访问该页；
  - `PTE_G`：页是否在所有地址空间中都有效；
  - `PTE_A`：页是否被访问过；
  - `PTE_D`：页是否被写过；
  - `PTE_RSW1` /  `PTE_RSW2`：保留给监视器软件使用的标志位。

# 物理内存管理

物理内存管理通过一个简单的空闲链表实现，包括初始化、分配和释放操作。该机制利用了自旋锁来确保多处理器环境下的同步性。

使用一个基于空闲链表的物理内存分配器来管理物理内存。空闲链表维护所有未分配的物理页，分配和释放物理内存页时通过操作这个链表来实现。

- 初始化时，所以物理页加入空闲页面链表。
- 每分配一个物理页，从空闲页面链表中取出一个物理页
- 释放物理页时，将物理页加入空闲页面链表

数据结构

```
struct {
  struct spinlock lock;
  struct run *freelist;
  uint64 npage;
} kmem;
```

物理块管理函数在 `kalloc.c` 中定义，包括：

- **`kinit()`**：初始化可用的物理地址空间：

  ```c
  void
  kinit()
  {
    initlock(&kmem.lock, "kmem");
    freerange(end, (void*)PHYSTOP);
  }
  ```

- **`kalloc()`**：返回一个可用的物理地址空间首地址：

  ```c
  void *
  kalloc(void)
  {
    struct run *r;
  
    acquire(&kmem.lock);
    r = kmem.freelist;
    if(r)
      kmem.freelist = r->next;
    release(&kmem.lock);
  
    if(r)
      memset((char*)r, 5, PGSIZE); // fill with junk
    return (void*)r;
  }
  ```

  >每次申请页面时，都会调用一次 `kalloc()`。`kalloc()`会从 `freelist` 中获取还未使用的物理页面并返回其首地址，对应的 `freelist` 已在 `kinit()` 中初始化过。

- **`kfree()`**：释放 `pa` 所指向的物理页面，并将其加入到可用的物理地址空间 `freelist` 中：

  ```c
  void
  kfree(void *pa)
  {
    struct run *r;
  
    if(((uint64)pa % PGSIZE) != 0 || (char*)pa < end || (uint64)pa >= PHYSTOP)
      panic("kfree");
  
    // Fill with junk to catch dangling refs.
    memset(pa, 1, PGSIZE);
  
    r = (struct run*)pa;
  
    acquire(&kmem.lock);
    r->next = kmem.freelist;
    kmem.freelist = r;
    release(&kmem.lock);
  }
  ```

# 虚拟内存管理

虚拟内存管理主要包括地址空间的映射、内存保护、页面分配和释放等功能。它使用分页机制将虚拟地址映射到物理地址。

但是需要注意有几个内核虚拟地址不是直接映射的：


- trampoline页面。它映射在虚拟地址空间的顶部。用户页表具有与此相同的映射。注意到，一个物理页面（包括trampoline代码）在内核的虚拟地址空间中映射了两次：一次在虚拟地址空间的顶部，一次是直接映射。    

- 内核堆栈页面。每个进程都有自己的内核堆栈，该堆栈被映射到较高的位置，因此xv6在其下方可以留下未映射的保护页。保护页的PTE无效（即，PTE\_V置0），因此，如果内核溢出内核堆栈，则很可能会导致异常，并且内核会出现紧急情况。如果没有保护页，溢出的堆栈将覆盖其他内核内存，从而导致错误的操作，系统很可能会崩溃。

 `vm.c` 文件中的 `kvminit()` 构建并初始化了**内核态的页表**。 `kvminit()` 执行以下操作：

- 对中断控制 `context` 及设备进行地址映射，其中所用到的宏定义被存储在 `memlayout.h` 文件中；
- 内核代码区、数据区和自由空间按直接映射方式映射，并设置相应访存权限；
- 将虚拟地址区域顶部即 `TRAMPOLINE` 指向的虚拟地址映射到 trap 出入口（`trampoline`）。

```c
void
kvminit()
{
  kernel_pagetable = (pagetable_t) kalloc();
  // printf("kernel_pagetable: %p\n", kernel_pagetable);

  memset(kernel_pagetable, 0, PGSIZE);
  
  //uint64 FAT32_SIZE=(FAT32_END-FAT32_START+PGSIZE-1)/PGSIZE*PGSIZE;
  //printf("fat32end:%p\n",FAT32_END);
  // file system
  //kvmmap(FAT32_START_V, FAT32_START, FAT32_SIZE, PTE_R | PTE_W);
  // uart registers
  kvmmap(UART0_V, UART0, PGSIZE, PTE_R | PTE_W);
  
  #if QEMU!=SIFIVE_U
  // virtio mmio disk interface
  kvmmap(VIRTIO0_V, VIRTIO0, PGSIZE, PTE_R | PTE_W);
  #endif
  // CLINT
  kvmmap(CLINT_V, CLINT, 0x10000, PTE_R | PTE_W);

  // PLIC
  kvmmap(PLIC_V, PLIC, 0x400000, PTE_R | PTE_W);

  #if QEMU==SIFIVE_U
  // GPIOHS
  kvmmap(GPIOHS_V, GPIOHS, 0x1000, PTE_R | PTE_W);

  // DMAC
  kvmmap(DMAC_V, DMAC, 0x1000, PTE_R | PTE_W);

  // GPIO
  // kvmmap(GPIO_V, GPIO, 0x1000, PTE_R | PTE_W);

  // SPI_SLAVE
  kvmmap(SPI_SLAVE_V, SPI_SLAVE, 0x1000, PTE_R | PTE_W);

  // FPIOA
  kvmmap(FPIOA_V, FPIOA, 0x1000, PTE_R | PTE_W);

  // SPI0
  kvmmap(SPI0_V, SPI0, 0x1000, PTE_R | PTE_W);

  // SPI1
  kvmmap(SPI1_V, SPI1, 0x1000, PTE_R | PTE_W);

  // SPI2
  kvmmap(SPI2_V, SPI2, 0x1000, PTE_R | PTE_W);

  // SYSCTL
  kvmmap(SYSCTL_V, SYSCTL, 0x1000, PTE_R | PTE_W);
  
  #endif
  
  // kvmmap(RUSTSBI_BASE, RUSTSBI_BASE, KERNBASE - RUSTSBI_BASE, PTE_R | PTE_X);
  // map kernel text executable and read-only.
  kvmmap(KERNBASE, KERNBASE, (uint64)etext - KERNBASE, PTE_R | PTE_X);
  // map kernel data and the physical RAM we'll make use of.
  kvmmap((uint64)etext, (uint64)etext, PHYSTOP - (uint64)etext, PTE_R | PTE_W);
  // map the trampoline for trap entry/exit to
  // the highest virtual address in the kernel.
  kvmmap(TRAMPOLINE, (uint64)trampoline, PGSIZE, PTE_R | PTE_X);

  #ifdef DEBUG
  printf("kvminit\n");
  #endif
}
```

在等待 `kvminit`() 构建好内核态使用的页表 `kernel_pagetable` 后，所有核可以调用 `kvminithart()` 将页表首地址载入到 `satp` 寄存器中。

```c
void
kvminithart()
{
  w_satp(MAKE_SATP(kernel_pagetable));
  // reg_info();
  sfence_vma();
  #ifdef DEBUG
  printf("kvminithart\n");
  #endif
}
```

##### 分配和释放用户虚拟内存

```
int allocuvm(pde_t *pgdir, uint oldsz, uint newsz) {
  char *mem;
  uint a;

  if (newsz >= KERNBASE)
    return 0;
  if (newsz < oldsz)
    return oldsz;

  a = PGROUNDUP(oldsz);
  for (; a < newsz; a += PGSIZE) {
    mem = kalloc();
    if (mem == 0) {
      cprintf("allocuvm out of memory\n");
      deallocuvm(pgdir, newsz, oldsz);
      return 0;
    }
    memset(mem, 0, PGSIZE);
    if (mappages(pgdir, (char*)a, PGSIZE, V2P(mem), PTE_W|PTE_U) < 0) {
      cprintf("allocuvm out of memory (2)\n");
      deallocuvm(pgdir, newsz, oldsz);
      return 0;
    }
  }
  return newsz;
}

int deallocuvm(pde_t *pgdir, uint oldsz, uint newsz) {
  pte_t *pte;
  uint a, pa;

  if (newsz >= oldsz)
    return oldsz;

  a = PGROUNDUP(newsz);
  for (; a  < oldsz; a += PGSIZE) {
    pte = walkpgdir(pgdir, (char*)a, 0);
    if (!pte)
      a += (NPTENTRIES - 1) * PGSIZE;
    else if ((*pte & PTE_P) != 0) {
      pa = PTE_ADDR(*pte);
      if (pa == 0)
        panic("kfree");
      char *v = P2V(pa);
      kfree(v);
      *pte = 0;
    }
  }
  return newsz;
}

```

## 页面置换策略：

页面置换策略采用的是简单的时钟算法（Clock algorithm），也称为最近未使用（Not Recently Used, NRU）算法的一种变体。

##### 工作原理：

1. 当需要选择一个页面进行替换时，时钟指针遍历环形链表。

2. 如果指针指向的页面的访问位为 0，则选择该页面进行替换，并将指针指向下一个页面。

3. 如果指针指向的页面的访问位为 1，则将访问位重置为 0，并将指针指向下一个页面，直到找到一个访问位为 0 的页面为止。