# 系统调用的过程

系统申请系统调用时，将依次执行以下函数：

- `uservec`
- `usertrap`
- 中断处理函数
- `usertrapret`
- `userret`

RISC-V 在中断响应时不会更换页表，所以我们需要操作系统在处理中断时将页表替换为内核页表，这可能导致缺页中断；其解决方案是添加一个 `trampoline` 页，这是一个特殊的页，包含 `uservec` 和 `userret` ，该页作为页表项存在于内核态和所有进程的页表中。所以无论什么进程请求系统调用时，都可以进入到统一的 `trampoline` 代码中。

由于 `stvec` 寄存器存储了 `uservec` 的地址，所以中断一开始的时候会进入 `uservec` 这个部分执行：

```assembly
uservec:    
  #
        # trap.c sets stvec to point here, so
        # traps from user space start here,
        # in supervisor mode, but with a
        # user page table.
        #
        # sscratch points to where the process's p->trapframe is
        # mapped into user space, at TRAPFRAME.
        #
        
        # swap a0 and sscratch
        # so that a0 is TRAPFRAME
        csrrw a0, sscratch, a0

        # save the user registers in TRAPFRAME
        sd ra, 40(a0)
        sd sp, 48(a0)
        sd gp, 56(a0)
        sd tp, 64(a0)
        sd t0, 72(a0)
        sd t1, 80(a0)
        sd t2, 88(a0)
        sd s0, 96(a0)
        sd s1, 104(a0)
        sd a1, 120(a0)
        sd a2, 128(a0)
        sd a3, 136(a0)
        sd a4, 144(a0)
        sd a5, 152(a0)
        sd a6, 160(a0)
        sd a7, 168(a0)
        sd s2, 176(a0)
        sd s3, 184(a0)
        sd s4, 192(a0)
        sd s5, 200(a0)
        sd s6, 208(a0)
        sd s7, 216(a0)
        sd s8, 224(a0)
        sd s9, 232(a0)
        sd s10, 240(a0)
        sd s11, 248(a0)
        sd t3, 256(a0)
        sd t4, 264(a0)
        sd t5, 272(a0)
        sd t6, 280(a0)

  # save the user a0 in p->trapframe->a0
        csrr t0, sscratch
        sd t0, 112(a0)

        # restore kernel stack pointer from p->trapframe->kernel_sp
        ld sp, 8(a0)

        # make tp hold the current hartid, from p->trapframe->kernel_hartid
        ld tp, 32(a0)

        # load the address of usertrap(), p->trapframe->kernel_trap
        ld t0, 16(a0)

        # restore kernel page table from p->trapframe->kernel_satp
        ld t1, 0(a0)
        csrw satp, t1
        sfence.vma zero, zero

        # a0 is no longer valid, since the kernel page
        # table does not specially map p->tf.

        # jump to usertrap(), which does not return
        jr t0
```

RISC-V 的寻址指令需要使用一个基址寄存器，在保存其他通用寄存器后需要 RISC-V 提供 `sscratch` 寄存器，这时可以把 a0 先暂时存储到 `sscratch` 中，再把 `a0` 从 `sscratch` 取出。

> 在这里，这个寄存器主要是存放了栈帧的首地址，新的栈帧就会存放在 `sscratch` 表示的地址中。在原文中提到，在返回到U态时，内核通过设置 `sscratch` 来制定下一次中断时栈帧的地址。

对于栈帧的处理同样需要页表。在操作系统中，每个进程我们都会申请一个 `trapframe` 页；对于用户态的地址空间中，这个页的虚拟地址永远指定在 `TRAPFRAME` 这个地址上；但是每个进程的页表是相互独立的，所以 `trapframe` 页的物理地址是不一样的。

这意味着对于内核态的代码，我们不能通过 `TRAPFRAME` 这个虚拟地址来访问进程的栈帧结构。所以在每个 `trapframe` 页的初始化过程中就需要保存好对应的物理地址：

```c
p->trapframe = (struct trapframe)kalloc();
```

此时 `kernel_trap` 已经保存到 `t0` 寄存器，那我们可以跳转到 `usertrap` 继续执行。

然后调用 `trap.c` 中的 `usertrap()` ，此时就已经进入内核态了。`usertrap()` 将执行以下操作：

- 修改 `stvec` 寄存器，因为对于在用户态和内核态发生中断请求，进入的中断程序是不一样的；
- 在 `trapframe` 里面在 `sepc` 寄存器保存中断断点；
- 对于不同的中断类型调用不同的处理函数。
  - 如果 trap 是 `syscall` ，调用  `syscall()` 即可；
  - 如果是设备故障，就先保存设备的编号；
  - 如果不是设备中断，就是指令的异常，直接退出；
  - 如果是时钟中断 `(which_dev==2)` ，就进入 `yield()` 进行处理。

```c
//
// handle an interrupt, exception, or system call from user space.
// called from trampoline.S
//
void
usertrap(void)
{
  //printf("run in usertrap\n");
  int which_dev = 0;

  if((r_sstatus() & SSTATUS_SPP) != 0)
    panic("usertrap: not from user mode");
  // send interrupts and exceptions to kerneltrap(),
  // since we're now in the kernel.
  w_stvec((uint64)kernelvec);

  //printf("user trap scause:%p\n",r_scause());
  struct proc *p = myproc();
  
  // save user program counter.
  p->trapframe->epc = r_sepc();
  
  if(r_scause() == 8){
    // system call
    if(p->killed)
      exit(-1);
    // sepc points to the ecall instruction,
    // but we want to return to the next instruction.
    p->trapframe->epc += 4;
    // an interrupt will change sstatus &c registers,
    // so don't enable until done with those registers.
    intr_on();
    syscall();
  } 
  else if((which_dev = devintr()) != 0){
    // ok
  } 
  else {
    printf("\nusertrap(): unexpected scause %p pid=%d %s\n", r_scause(), p->pid, p->name);
    printf("            sepc=%p stval=%p\n", r_sepc(), r_stval());
    // trapframedump(p->trapframe);
    p->killed = 1;
  }

  if(p->killed)
    exit(-1);

  // give up the CPU if this is a timer interrupt.
  if(which_dev == 2)
    yield();

  usertrapret();
}
```

#### syscall()处理

在 `usys.pl` 脚本中，我们编写了系统调用的渲染脚本。

```perl
sub entry {
    my $name = shift;
    print ".global $name\n";
    print "${name}:\n";
    print " li a7, SYS_${name}\n";
    print " ecall\n";
    print " ret\n";
}
```

系统调用号保存在 `a7` 寄存器中，读取这个值后，通过调用函数指针数组的值可以跳转至对应函数。

#### 中断返回

中断返回首先调用 `usertrapret()` ，该函数执行以下操作：

- **写入 `stvec`**：将 `stvec` 从 `kernelvec` 修改为`uservec` ；

- **处理 `trapframe`**：保存内核页表地址、内核栈、 `usertrap` 地址和 CPU 核号；

- **更改 `sstatus`**；

- **将断点地址写入 `sepc`** 。

```c
void
usertrapret(void)
{
  struct proc *p = myproc();

  // we're about to switch the destination of traps from
  // kerneltrap() to usertrap(), so turn off interrupts until
  // we're back in user space, where usertrap() is correct.
  intr_off();

  // send syscalls, interrupts, and exceptions to trampoline.S
  w_stvec(TRAMPOLINE + (uservec - trampoline));

  // set up trapframe values that uservec will need when
  // the process next re-enters the kernel.
  p->trapframe->kernel_satp = r_satp();         // kernel page table
  p->trapframe->kernel_sp = p->kstack + PGSIZE; // process's kernel stack
  p->trapframe->kernel_trap = (uint64)usertrap;
  p->trapframe->kernel_hartid = r_tp();         // hartid for cpuid()

  // set up the registers that trampoline.S's sret will use
  // to get to user space.
  
  // set S Previous Privilege mode to User.
  unsigned long x = r_sstatus();
  x &= ~SSTATUS_SPP; // clear SPP to 0 for user mode
  x |= SSTATUS_SPIE; // enable interrupts in user mode
  w_sstatus(x);

  // set S Exception Program Counter to the saved user pc.
  w_sepc(p->trapframe->epc);

  // tell trampoline.S the user page table to switch to.
  uint64 satp = MAKE_SATP(p->pagetable);

  // jump to trampoline.S at the top of memory, which 
  // switches to the user page table, restores user registers,
  // and switches to user mode with sret.
  uint64 fn = TRAMPOLINE + (userret - trampoline);
  ((void (*)(uint64,uint64))fn)(TRAPFRAME, satp);
}
```

进入 `userret` 前先传递参数：`TRAPFRAME` 作为第一个参数，用户态页表的地址作为第二个参数：

- 加载用户态页表；
- 栈帧中存储的寄存器值全部加载到真实的寄存器中；
- 栈帧头部的虚拟地址保存到 `sscratch` 寄存器；

最后执行 `sret` ，把 `sepc` 寄存器的内容传递给 PC，转换为U态，中断结束。

```asm
userret:
        # userret(TRAPFRAME, pagetable)
        # switch from kernel to user.
        # usertrapret() calls here.
        # a0: TRAPFRAME, in user page table.
        # a1: user page table, for satp.

        # switch to the user page table.
        csrw satp, a1
        sfence.vma zero, zero

        # put the saved user a0 in sscratch, so we
        # can swap it with our a0 (TRAPFRAME) in the last step.
        ld t0, 112(a0)
        csrw sscratch, t0

        # restore all but a0 from TRAPFRAME
        ld ra, 40(a0)
        ld sp, 48(a0)
        ld gp, 56(a0)
        ld tp, 64(a0)
        ld t0, 72(a0)
        ld t1, 80(a0)
        ld t2, 88(a0)
        ld s0, 96(a0)
        ld s1, 104(a0)
        ld a1, 120(a0)
        ld a2, 128(a0)
        ld a3, 136(a0)
        ld a4, 144(a0)
        ld a5, 152(a0)
        ld a6, 160(a0)
        ld a7, 168(a0)
        ld s2, 176(a0)
        ld s3, 184(a0)
        ld s4, 192(a0)
        ld s5, 200(a0)
        ld s6, 208(a0)
        ld s7, 216(a0)
        ld s8, 224(a0)
        ld s9, 232(a0)
        ld s10, 240(a0)
        ld s11, 248(a0)
        ld t3, 256(a0)
        ld t4, 264(a0)
        ld t5, 272(a0)
        ld t6, 280(a0)

  # restore user a0, and save TRAPFRAME in sscratch
        csrrw a0, sscratch, a0
        
        # return to user mode and user pc.
        # usertrapret() set up sstatus and sepc.
        sret
```

#### 调用系统函数

下面是一个标准的系统调用示例， `a0` 至 `a6` 寄存器存储系统调用需要的参数， `a7` 传递系统调用号，传递完参数后就执行`ecall`。执行 `ecall` 指令会将状态调整为 S 态，然后执行 `uservec()` ，随后进入 `trap()` 处理函数：

```asm
#include "syscall.h"

# exec(init, argv)
.globl start
start:
        la a0, init
        la a1, argv
        li a7, SYS_exec
        ecall

# for(;;) exit();
exit:
        li a7, SYS_exit
        ecall
        jal exit

# char init[] = "/init\0";
init:
  .string "/init\0"

# char *argv[] = { init, 0 };
.p2align 2
argv:
  .long init
  .long 0
```

返回时会将返回值传递给 `a0` 寄存器。

```c
static uint64 (*syscalls[])(void) = {
[SYS_fork]    sys_fork,
[SYS_exit]    sys_exit,
[SYS_wait]    sys_wait,
[SYS_pipe]    sys_pipe,
[SYS_read]    sys_read,
[SYS_kill]    sys_kill,
[SYS_exec]    sys_exec,
[SYS_fstat]   sys_fstat,
[SYS_chdir]   sys_chdir,
[SYS_dup]     sys_dup,
[SYS_getpid]  sys_getpid,
[SYS_sbrk]    sys_sbrk,
[SYS_sleep]   sys_sleep,
[SYS_uptime]  sys_uptime,
[SYS_open]    sys_open,
[SYS_write]   sys_write,
[SYS_mknod]   sys_mknod,
[SYS_unlink]  sys_unlink,
[SYS_link]    sys_link,
[SYS_mkdir]   sys_mkdir,
[SYS_close]   sys_close,
};
//系统调用号,系统调用函数
void
syscall(void)
{
  int num;
  struct proc *p = myproc();

  num = p->trapframe->a7;
  if(num > 0 && num < NELEM(syscalls) && syscalls[num]) {
    p->trapframe->a0 = syscalls[num]();
  } else {
    printf("%d %s: unknown sys call %d\n",
            p->pid, p->name, num);
    p->trapframe->a0 = -1;
  }
}
```

#### 系统调用的参数

系统调用会传递参数进入，对于 RISC-V 而言，朴素的思想就是把参数传递到寄存器中，然后系统调用函数读取存储在寄存器中的数据。

- 对于直接传递的参数，可以直接读取；
- 对于指针参数，需要额外处理以确保安全性和处理内核态与用户态页表的差异。
  - 我们无法确定程序是否是友好的，用户程序可能通过传递地址来修改内核的内存，导致不安全情况的发生；
  - xv6 的内核态和用户态页表不同。

xv6 的做法是：对于获得字符串的函数 `argstr()` ，去构建一个新的函数 `fetchstr()` 去安全地获得数据，这个函数就会调用 `copyinstr()` 。

```c
int
copyinstr(pagetable_t pagetable, char *dst, uint64 srcva, uint64 max)
{
  uint64 n, va0, pa0;
  int got_null = 0;

  while(got_null == 0 && max > 0){
    va0 = PGROUNDDOWN(srcva);
    pa0 = walkaddr(pagetable, va0);
    if(pa0 == 0)
      return -1;
    n = PGSIZE - (srcva - va0);
    if(n > max)
      n = max;

    char *p = (char *) (pa0 + (srcva - va0));
    while(n > 0){
      if(*p == '\0'){
        *dst = '\0';
        got_null = 1;
        break;
      } else {
        *dst = *p;
      }
      --n;
      --max;
      p++;
      dst++;
    }

    srcva = va0 + PGSIZE;
  }
  if(got_null){
    return 0;
  } else {
    return -1;
  }
}
```

`copyinstr()` 会从 `pagetable` 页表对应的虚拟地址 `srcva` 处拷贝 `max` 字节的元素到内核页表的 `dst` 处，做法就是调用 `walkaddr` 来找到 `pagetable` 中 `srcva` 对应的物理地址，然后从该物理地址中拷贝字节到 `dst` 内。

总体的思路就是，找到这个用户态虚拟地址对应的物理地址，取出这个物理地址上的元素，然后拷贝到 `dst` 内。