# 一、文件管理原理

## 文件描述符与文件对象

**文件描述符机制**：

- 每个进程维护一个文件描述符表（`fdtable`），通过文件描述符（`fd`）指向文件对象（`struct file`）。
- 文件对象包含文件的元数据、访问模式（可读/可写）和文件偏移量等信息。

**文件对象与目录项关联**：

- 文件对象通过目录项（`struct dirent`）与文件系统中的实际文件关联。
- 目录项记录了文件的名称、类型（文件/目录）及其在文件系统中的位置。

## **文件打开与创建**

**文件打开（openat 系列调用）**：

- 用户通过`openat`指定文件路径及访问模式，系统根据路径找到文件对应的目录项并创建文件对象。
- 如果指定了`O_CREATE`，文件不存在时会自动创建文件。

**路径解析与权限检查**：

- 通过递归解析路径中的每一级目录项，最终定位目标文件或目录。
- 对目标文件检查访问模式是否符合要求（如是否为目录、是否可读写）。

## **文件读写**

**文件偏移量管理**：

- 每个文件对象包含一个独立的偏移量，表示当前的读写位置。
- 通过`read`和`write`调用读取或写入文件内容，并根据访问模式自动更新偏移量。

**文件内容的加载与更新**：

- 从文件系统中读取文件内容到内存，通过文件对象的偏移量操作数据。

- 写入时可选择`O_APPEND`模式，将数据追

## **目录操作**

**目录项读取**：

- 用户打开目录读取目录内容，返回目录项列表（包括文件名、类型等信息）。
- 系统遍历目录项，将对应的内容返回到用户缓冲区。

**目录项创建与删除**：

- 通过`mkdir`等函数创建目录，将目录项插入到父目录中。
- 通过`remove`等函数删除文件或目录，确保目标为空目录或普通文件，避免误删非空目录。

## **文件系统一致性保障**

**锁机制**：

- 文件和目录项在操作时使用锁（如`elock`）保护，确保多进程并发操作的正确性。

**引用计数管理**：

- 文件对象和目录项都维护引用计数，只有当引用计数降为零时，才释放相应资源。

**延迟写入与同步**：

- 文件内容的修改通常先在内存中缓存，最终通过文件系统的同步机制写入磁盘，保证一致性和效率。



# 二、扩展的功能

新增实现了dup2，openat，getdents64，unlinkat，mkdirat系统调用和用户函数，以及chmod，chown

## 1.dup2 系统调用

**功能：**复制已有的文件描述符到指定的新文件描述符。如果新文件描述符已被占用，则先关闭它。

### **实现流程**

1. 参数解析：获取源文件描述符对应的文件对象。获取目标文件描述符。
2. 分配新文件描述符：调用 `fdalloc_new` 将源文件对象绑定到目标文件描述符。
3. 增加引用计数：调用 `filedup` 增加文件对象的引用计数，避免重复释放。
4. **返回目标文件描述符**。

代码

```
uint64 sys_dup2(void) {
    struct file *f;
    int new;
    int fd;
    // 解析源文件描述符和目标文件描述符
    if (argfd(0, 0, &f) < 0 || argint(1, &new) < 0) {
        return -1;  // 参数解析失败
    }
    // 分配新文件描述符
    if ((fd = fdalloc_new(f, new)) < 0) {
        return -1;  // 分配失败
    }
    // 增加文件对象引用计数
    filedup(f);
    return fd;
}
```

调用 `fdalloc_new` 分配目标描述符。

使用 `filedup` 增加引用计数，确保文件资源正确管理。

```
//如果目标文件描述符 new 已被占用,关闭目标文件描述符 
//分配目标文件描述符 new,将 old 文件描述符复制到目标位置
static int
fdalloc_new(struct file *f, int new)
{
    struct proc *p = myproc(); // 获取当前进程结构

    // 检查文件描述符范围是否合法
    if (new < 0 || new >= NOFILE) {
        return -1; // 文件描述符超出范围
    }

    // 如果目标描述符已经被占用，先关闭它
    if (p->ofile[new] != 0) {
        fileclose(p->ofile[new]); // 关闭旧文件
        p->ofile[new] = 0;        // 清空描述符表对应的条目
    }

    // 分配新的文件描述符
    p->ofile[new] = f; // 将目标文件描述符指向文件 `f`
    return new;        // 返回目标文件描述符
}
```

## 2.openat 系统调用

**功能：**在指定路径打开文件或目录，支持相对路径操作。可选创建文件，并设置访问模式和权限。

### **实现流程**

1. 参数解析：获取目录文件描述符、路径、标志和模式。
2. 路径解析：调用 `ename_with_fd` 根据文件描述符和路径获取目录项。如果设置了 `O_CREATE` 且文件不存在，调用 `create_with_fd` 创建文件。
3. 权限检查：检查是否符合 `O_DIRECTORY` 和 `O_RDONLY` 等标志要求。
4. 分配文件结构：分配文件对象，初始化访问模式、偏移量和文件类型。
5. **返回文件描述符**。

代码：
  openat的参数:
  flags表示打开文件的方式，与open中的omode一致，是文件打开操作的模式标志
  mode则表示文件的属性,新建文件时需指定，是文件/目录的属性标志

```

uint64 sys_openat(void) {
    char path[FAT32_MAX_PATH];
    int fd, flags, mode;
    struct dirent *entry;

    if (argint(0, &fd) < 0 || argstr(1, path, FAT32_MAX_PATH) < 0 || argint(2, &flags) < 0 || argint(3, &mode)) {
        return -1;  // 参数解析失败
    }

    if ((entry = ename_with_fd(fd, path)) == NULL) {
        if (flags & O_CREATE) {  // 文件不存在时创建
            if ((entry = create_with_fd(fd, path, T_FILE, mode)) == NULL) {
                return -1;
            }
        } else {
            return -1;  // 文件不存在且未指定 O_CREATE
        }
    }

    // 检查目录和权限标志
    if (!(entry->attribute & ATTR_DIRECTORY) && (flags & O_DIRECTORY)) {
        eunlock(entry);
        eput(entry);
        return -1;
    }

    // 分配文件描述符并初始化文件结构
    struct file *file;
    if ((file = filealloc()) == NULL || (fd = fdalloc(file)) < 0) {
        if (file) fileclose(file);
        eunlock(entry);
        eput(entry);
        return -1;
    }

    // 设置文件的偏移量和属性
    file->type = FD_ENTRY;
    file->off = (flags & O_APPEND) ? entry->file_size : 0;
    file->readable = !(flags & O_WRONLY);
    file->writable = (flags & O_WRONLY) || (flags & O_RDWR);

    eunlock(entry);
    return fd;
}
```

支持相对路径解析。

支持创建文件 (`O_CREATE`) 和清空文件 (`O_TRUNC`)。

设置文件访问模式和偏移量。

### 辅助函数

openat调用的其他函数，路径`kernel/fat32.c`

```
create_with_fd(int fd, char *path, short type, int mode)
ename_with_fd(int fd, char *path)
```

ename_with_fd：

- 调用 `lookup_path_with_fd` 函数，并设置 `parent=0`，以获取目标文件或目录的目录项。
- 提供了从路径中解析并定位目标目录或文件的功能。

create_with_fd：

- 根据指定的路径和文件类型，在路径中创建新文件或目录。
- 如果路径不存在，会递归地创建父目录。
- 调用 `ealloc` 为目标文件或目录分配目录项。
- 根据文件类型和模式设置文件属性。
- 返回新创建的文件或目录的目录项指针

lookup_path_with_fd：

- 根据文件描述符 `fd` 和路径 `path` 解析目标文件或目录的目录项（`dirent`），并返回其指针。
- 支持从当前工作目录（`cwd`）、根目录（`root`）或指定的文件描述符所指目录开始解析。
- 逐层解析路径元素，确保路径合法并且每一级都为目录。
- 如果 `parent` 为 1，则返回目标文件或目录的父目录项。
- 如果解析失败或路径非法，返回 `NULL`。



## 3.getdents64 系统调用

**功能:**读取目录内容，将目录项以 `linux_dirent64` 格式返回给用户态。

### **实现流程**

1. 参数解析：获取目录文件描述符和用户缓冲区信息。
2. 权限检查：检查文件是否为目录，是否可读。
3. 遍历目录项：调用 `enext` 遍历目录内容，跳过空目录项。填充 `linux_dirent64` 结构，将信息拷贝到用户空间。
4. 返回结果：返回成功读取的字节数，或返回错误。

代码：

```

uint64 sys_getdents64(void)
{
    struct proc *p = myproc();  // 获取当前进程
    struct file *f;             // 文件指针，用于操作目录
    int fd;                     // 文件描述符
    uint64 buf, len;            // 用户提供的缓冲区地址和缓冲区长度

    // 解析用户传入的参数：
    // 参数 0: 文件描述符
    // 参数 1: 用户缓冲区地址
    // 参数 2: 用户缓冲区大小
    if (argfd(0, &fd, &f) < 0 || argaddr(1, &buf) < 0 || argaddr(2, &len) < 0) {
        return -1;  // 如果参数解析失败，则返回 -1
    }

    // 检查文件是否可读，以及是否是目录
    if (f->readable == 0 || !(f->ep->attribute & ATTR_DIRECTORY)) {
        return -1;  // 如果不是可读的目录文件，则返回 -1
    }

    int err;                       // 用于存储读取目录项时的错误状态
    struct dirent dste;            // 文件系统内部的目录项结构
    dirent64 dir64;                // 用于返回给用户的 linux_dirent64 结构
    int copysize = 0;              // 已经复制到用户缓冲区的字节数
    int count = 0;                 // 当前目录项占用的条目数量

    elock(f->ep);                  // 加锁，确保读取目录项时的一致性

    // 遍历目录内容
    while (1) {
        dir64.d_off = f->off;      // 设置当前目录项的偏移量
        // 跳过空目录项
        while ((err = enext(f->ep, &dste, f->off, &count)) == 0) {
            f->off += count * 32;  // 如果是空目录项，跳过，更新文件偏移量
        }

        if (err == -1) {           // 如果到达目录末尾，退出循环
            break;
        }

        f->off += count * 32;      // 读取非空目录项后，更新文件偏移量

        // 填充 linux_dirent64 结构
        dir64.d_type = (dste.attribute & ATTR_DIRECTORY) ? T_DIR : T_FILE;  // 判断目录项类型
        dir64.d_ino = 0;           // 暂时未用，设置为 0
        strncpy(dir64.d_name, dste.filename, sizeof(dste.filename));  // 拷贝文件名
        dir64.d_reclen = strlen(dir64.d_name) + sizeof(dirent64) - sizeof(dir64.d_name) + 1;  // 计算记录长度

        // 检查缓冲区是否足够存储当前目录项
        if (dir64.d_reclen > len) {
            break;  // 如果缓冲区不足，停止读取
        }

        int copySize = dir64.d_reclen;  // 记录当前目录项的大小

        // 将目录项从内核空间复制到用户空间
        if (copyout(p->pagetable, buf, (char *)&dir64, copySize) < 0) {
            eunlock(f->ep);  // 如果复制失败，解锁并返回错误
            return -1;
        }

        buf += copySize;       // 更新用户缓冲区指针
        len -= copySize;       // 更新剩余缓冲区大小
        copysize += copySize;  // 累计已复制的总字节数
    }

    eunlock(f->ep);  // 解锁目录项
    return copysize; // 返回成功写入用户缓冲区的字节数
}
```

使用 `linux_dirent64` 格式返回目录项。

遍历时跳过空目录项，减少无效数据传输。

```
//linux的目录项结构体
typedef struct linux_dirent64 {
    uint64 d_ino; // 索引结点号
    uint64 d_off; // 到下一个dirent的偏移
    unsigned short d_reclen; // 当前dirent的长度
    unsigned char d_type; // 文件类型
    char d_name[FAT32_MAX_FILENAME + 1]; //文件名
}__attribute__((packed, aligned(4))) dirent64;
```



## 4.unlinkat 系统调用

**功能：** 用于删除文件或空目录。

### **实现流程**

1. 参数解析：获取文件路径并处理尾部斜杠。
2. 权限检查：禁止删除当前目录 `.` 或父目录 `..`。检查目录是否为空。
3. 删除目录项：锁定父目录，调用 `eremove` 删除文件。解锁并释放引用。

代码：

```
uint64 
sys_unlinkat(void)
{
    char path[FAT32_MAX_PATH];      // 用于存储路径参数
    struct dirent *ep;              // 指向目标文件的目录项
    char name[FAT32_MAX_FILENAME];  // 保存路径中的文件名
    struct dirent *parent;          // 指向父目录的目录项
    int dirfd;                      // 文件描述符

    // 解析用户参数
    if (argint(0, &dirfd) < 0 || argstr(1, path, FAT32_MAX_PATH) < 0) {
        return -1;  // 参数解析失败
    }

    // 解析路径并获取父目录和文件名
    if ((parent = enameparent_with_fd(dirfd, path, name)) == NULL) {
        return -1;  // 父目录解析失败，返回 -1
    }

    // 加锁父目录
    elock(parent);

    // 查找目标目录项
    if ((ep = dirlookup(parent, name, 0)) == NULL) {
        eunlock(parent);
        eput(parent);
        return -1;  // 文件不存在，返回 -1
    }

    // 锁定目标文件
    elock(ep);

    // 检查目标是否是目录
    if (ep->attribute & ATTR_DIRECTORY) {
        // 如果是目录，返回错误
        eunlock(ep);
        eput(ep);
        eunlock(parent);
        eput(parent);
        return -1;  // 目标是目录，返回 -1
    }

    // 删除目录项
    eremove(ep);

    // 解锁并释放引用
    eunlock(ep);
    eput(ep);
    eunlock(parent);
    eput(parent);

    return 0;  // 删除成功，返回 0
}

```



## 5.mkdirat 系统调用

**功能：**`mkdirat` 用于在指定路径创建目录。支持相对路径操作，通过文件描述符 (`dirfd`) 指定相对路径起点。

### **实现流程**

1. 参数解析：获取路径参数。
2. 创建目录项：调用 `create` 创建目录项，设置类型为目录。
3. 返回结果：返回 0 表示成功，-1 表示失败。

### 参数说明

1. dirfd：文件描述符，指定路径解析的起点。

    特殊值：

    - `AT_FDCWD`：表示从当前工作目录解析路径。

    - 如果路径是绝对路径，则忽略 `dirfd`。

2. pathname：要删除的目标文件路径。可以是绝对路径或相对路径。

3. flags：标志位

代码：

```
uint64 
sys_mkdirat(void)
{
    char path[FAT32_MAX_PATH];      // 用于存储路径参数
    struct dirent *parent;          // 父目录的目录项
    struct dirent *ep;              // 指向新建的目录项
    char name[FAT32_MAX_FILENAME];  // 保存路径中的目录名
    int dirfd, mode;                // 文件描述符和模式

    // 参数解析
    if (argint(0, &dirfd) < 0 || argstr(1, path, FAT32_MAX_PATH) < 0 || argint(2, &mode) < 0) {
        return -1;  // 参数解析失败，返回 -1
    }

    // 解析路径并获取父目录和目录名
    if ((parent = enameparent_with_fd(dirfd, path, name)) == NULL) {
        return -1;  // 父目录解析失败，返回 -1
    }

    // 加锁父目录
    elock(parent);

    // 检查是否同名目录或文件已存在
    if ((ep = dirlookup(parent, name, 0)) != NULL) {
        eunlock(parent);
        eput(parent);
        eput(ep);
        return -1;  // 如果同名目录或文件已存在，返回 -1
    }

    // 创建新目录
    if ((ep = ealloc(parent, name, ATTR_DIRECTORY)) == NULL) {
        eunlock(parent);
        eput(parent);
        return -1;  // 如果分配失败，返回 -1
    }

    // 设置新目录的模式（目前仅存储模式值，不具体处理权限）
    ep->mode = mode;

    // 解锁父目录并释放其引用计数
    eunlock(parent);
    eput(parent);

    // 解锁新建目录项并释放其引用计数
    eunlock(ep);
    eput(ep);

    return 0;  // 成功返回 0
}

```

满足了基础的 `unlinkat` 功能要求，可以进一步扩展。



## 6.chmod,chown

这两个函数用于文件权限管理，虽然实现了函数，但是由于xv6-riscv-k210项目使用的文件系统是FAT32,它不支持文件权限的存储，所以没有进行真正的应用，想要进行权限管理，可以考虑通过内存映射表建立权限的映射关系。

## sys_chmod 系统调用

**功能：**修改文件的权限模式（`mode`）

**参数**：

- **`path`**：文件路径（字符串）。
- **`mode`**：新的权限模式（整数，八进制，范围 0000-0777）。

**限制**：

​	只有文件的拥有者或超级用户（`uid == 0`）才能修改权限。

### **实现流程**

1. **参数解析**：从用户空间获取两个参数：文件路径 `path`。新的权限模式 `mode`。
2. **参数合法性检查**：检查 `mode` 是否在有效范围内（`0000-0777`）。
3. **文件查找**：调用 `ename(path)` 函数查找文件对应的目录项 `dirent`。
4. **权限检查**：检查当前用户（`myproc()->uid`）是否是文件的拥有者或超级用户（`uid == 0`）。如果无权限，返回错误。
5. **权限修改**：更新文件的权限模式为 `mode`。调用 `eupdate(ep)` 将更改写回磁盘。
6. **资源释放**：解锁文件 `eunlock(ep)`。释放目录项引用 `eput(ep)`。
7. **返回结果**：成功返回 `0`。失败返回 `-1`。

代码实现：

```
int sys_chmod(void) {
  char path[FAT32_MAX_PATH];
  int mode;
  struct dirent *ep;

  // 从用户空间获取文件路径和权限
  if (argstr(0, path, FAT32_MAX_PATH) < 0 || argint(1, &mode) < 0)
    return -1;

  // 检查权限值是否合理
  if (mode < 0 || mode > 0777) {
    return -1; // 无效的权限值
  }


  if ((ep = ename(path)) == NULL) { // 查找文件
     
    return -1;
  }
  elock(ep);

  // 检查是否有权限修改文件（只有文件拥有者或 root 可以修改权限）
  if (myproc()->uid != ep->uid && myproc()->uid != 0) {
    eunlock(ep);
    eput(ep);
     
    return -1; // 当前用户无权限修改
  }

  // 更新文件权限
  ep->mode = (short)mode;
  eupdate(ep); // 写回更改到磁盘
  eunlock(ep);
  eput(ep);


  return 0; // 修改成功
}
```



## sys_chown 系统调用

- **功能**：修改文件的拥有者（`uid`）和所属组（`gid`）。
- 参数：
    - **`path`**：文件路径（字符串）。
    - **`uid`**：新的拥有者用户 ID（整数，范围 0-65535）。
    - **`gid`**：新的所属组 ID（整数，范围 0-65535）。
- 限制：只有文件的拥有者或超级用户（`uid == 0`）才能修改拥有者和组。

### **`sys_chown` 实现流程**

1. **参数解析**：从用户空间获取三个参数：文件路径 `path`。新的用户 ID `uid`。新的组 ID `gid`。
2. **参数合法性检查**：检查 `uid` 和 `gid` 是否在有效范围内。
3. **文件查找**：调用 `ename(path)` 函数查找文件对应的目录项 `dirent`。
4. **权限检查**：检查当前用户（`myproc()->uid`）是否是文件的拥有者或超级用户（`uid == 0`）。如果无权限，返回错误。
5. **拥有者和组修改**：更新文件的拥有者为 `uid`，更新文件的所属组为 `gid`，调用 `eupdate(ep)` 将更改写回磁盘。
6. **资源释放**：解锁文件 `eunlock(ep)`。释放目录项引用 `eput(ep)`。
7. **返回结果**：成功返回 `0`。失败返回 `-1`。



代码实现:

```

/* 
  sys_chown 用于修改文件的拥有者（uid）和所属组（gid）
  接受文件路径、用户 ID 和组 ID 作为参数
 */
int sys_chown(void) {
  char path[FAT32_MAX_PATH];
  int uid, gid;
  struct dirent *ep;

  // 从用户空间获取文件路径、用户 ID 和组 ID
  if (argstr(0, path, FAT32_MAX_PATH) < 0 || argint(1, &uid) < 0 || argint(2, &gid) < 0)
    return -1;

  // 检查用户 ID 和组 ID 是否合理
  if (uid < 0 || uid > 65535 || gid < 0 || gid > 65535) {
    return -1; // 无效的 uid 或 gid
  }

  if ((ep = ename(path)) == NULL) { // 查找文件
     
    return -1;
  }
  elock(ep);

  // 检查是否有权限修改文件（只有文件拥有者或 root 可以修改拥有者和组）
  if (myproc()->uid != ep->uid && myproc()->uid != 0) {
    eunlock(ep);
    eput(ep);
     
    return -1; // 当前用户无权限修改
  }

  // 更新文件拥有者和所属组
  ep->uid = (short)uid;
  ep->gid = (short)gid;
  eupdate(ep); // 写回更改到磁盘
  eunlock(ep);
  eput(ep);


  return 0; // 修改成功
}
```



# 三、总结

对文件管理的功能进行了扩展，实现符合POSIX 标准的功能，基本包括以下特点

**路径解析**：所有系统调用均支持绝对路径和相对路径解析，通过文件描述符 `dirfd` 实现灵活性。

**并发安全**：通过锁机制（如 `elock` 和 `eunlock`）保证资源管理的线程安全。

**错误处理**：每个系统调用均对路径解析失败、权限错误和资源不足等情况进行了细致处理。

**资源管理**：通过引用计数 (`eput`) 和锁释放，确保资源不泄露或冲突。