# tab补全

习惯了tab补全功能，总是下意识地按tab键，但是在xv6上并没有实现tab补全，导致使用起来很不顺手，于是尝试添加tab补全功能。



首先让我们理一下xv6是如何获取用户的输入信息的。

tarp.c中可以看到当中断为UART时，调用sbi_console_getchar()获取用户输入。然后进入consoleintr(c);

```c
int devintr(void) {
	uint64 scause = r_scause();

	#ifdef QEMU 
	// handle external interrupt 
	if ((0x8000000000000000L & scause) && 9 == (scause & 0xff)) 
	#else 
	// on k210, supervisor software interrupt is used 
	// in alternative to supervisor external interrupt, 
	// which is not available on k210. 
	if (0x8000000000000001L == scause && 9 == r_stval()) 
	#endif 
	{
		int irq = plic_claim();
		if (UART_IRQ == irq) {
			// keyboard input 
			int c = sbi_console_getchar();
      
			if (-1 != c) {
				consoleintr(c);
			}
		}
		else if (DISK_IRQ == irq) {
			disk_intr();
		}
		else if (irq) {
			printf("unexpected interrupt irq = %d\n", irq);
		}

		if (irq) { plic_complete(irq);}

		#ifndef QEMU 
		w_sip(r_sip() & ~2);    // clear pending bit
		sbi_set_mie();
		#endif 

		return 1;
	}
	else if (0x8000000000000005L == scause) {
		timer_tick();
		return 2;
	}
	else { return 0;}
}
```

然后再看consoleintr函数：

根据用户输入的不同的c，做出相应的处理，在default中，会默认调用consputc(c)将用户输入的字符回显，即每次我们按下键盘都能看到输入的内容。所以我们可以在这里对'\t'，也就是tab键做特殊处理，也就是不回显。

```c
void
consoleintr(int c)
{
  acquire(&cons.lock);

  switch(c){
  case C('P'):  // Print process list.
    procdump();
    break;
  case C('U'):  // Kill line.
    while(cons.e != cons.w &&
          cons.buf[(cons.e-1) % INPUT_BUF] != '\n'){
      cons.e--;
      consputc(BACKSPACE);
    }
    break;
  case C('H'): // Backspace
  case '\x7f':
    if(cons.e != cons.w){
      cons.e--;
      consputc(BACKSPACE);
    }
    break;
  default:
    if(c != 0 && cons.e-cons.r < INPUT_BUF){
      #ifndef QEMU
      if (c == '\r') break;     // on k210, "enter" will input \n and \r
      #else
      c = (c == '\r') ? '\n' : c;
      #endif

      if(c == '\t')		//特殊处理
      {
        cons.buf[cons.e++ % INPUT_BUF] = c;
        cons.w = cons.e;
        tab_flag = 1;
        wakeup(&cons.r);
        break;
      }
      // echo back to the user.
      consputc(c);

      // store for consumption by consoleread().
      cons.buf[cons.e++ % INPUT_BUF] = c;

      if(c == '\n' || c == C('D') || cons.e == cons.r+INPUT_BUF){
        // wake up consoleread() if a whole line (or end-of-file)
        // has arrived.
        cons.w = cons.e;
        wakeup(&cons.r);
      }
    }
    break;
  }
  
  release(&cons.lock);
}
```

那shell是如何获取用户输入的命令的呢，我们现在看sh.c:

main函数里会调用getcmd获取命令,然后在getcmd中调用read阻塞等待用户输入，这里read函数是从console.c中的consoleread函数获取，详细可去看consoleread函数。所以，从read用户输入的命令，那我们可以在这里对't'做补全的功能就行了。

```c
int
main(void)
{
  .......
  while(getcmd(buf, sizeof(buf)) >= 0){
    replace(buf);
    if(buf[0] == 'c' && buf[1] == 'd' && buf[2] == ' '){
      // Chdir must be called by the parent, not the child.
      buf[strlen(buf)-1] = 0;  // chop \n
      if(chdir(buf+3) < 0)
        fprintf(2, "cannot cd %s\n", buf+3);
      getcwd(mycwd);
    }
  .......
  exit(0);
}

int
getcmd(char *buf, int nbuf)
{
  fprintf(2, "-> %s $ ", mycwd);
  memset(buf, 0, nbuf);


  int i, cc;
  char c;

  for(i=0; i+1 < nbuf; ){
    cc = read(0, &c, 1);
    if(cc < 1)
      break;
    if(c == '\t')
    {
      i += tab_completion(buf);		//tab补全
    }
    else
    {
      buf[i++] = c;
    }
    
    if(c == '\n' || c == '\r')
      break;
  }
  buf[i] = '\0';

  if(buf[0] == 0) // EOF
    return -1;
  return 0;
}

```

tab_completion:

其中cmdbuf是用户输入的命令，如cd a，这时调用find_last_word函数，将会返回a，然后，这里仿照find.c实现，查找当前目录下的所有文件名，然后和用户输入的内容进行比对，若成功则补全。

```c
int tab_completion(char *cmdbuf) {
  char *last_word = find_last_word(cmdbuf);
  //printf("\ntrying tab:%s\n", last_word);

  int fd;
  struct stat st;
  const char* path = ".";


  if ((fd = open(path, O_RDONLY)) < 0)
  {
    fprintf(2, "find: cannot open %s\n", path);
    return;
  }

  if (fstat(fd, &st) < 0)
  {
    fprintf(2, "find: cannot stat %s\n", path);
    close(fd);
    return;
  }

  int added_length = 0;

  if (st.type != T_DIR)
  {
    close(fd);
    return;
  }

  //这里使用buf来存储path，因为下面char *p又指向了path,虽然定义的path是const char* path = "."，
  //这表示path指向的"."是不可变的，但是p这个指针指向了同一块地址，所以path的内容是可以改变的，所以这就是为什么const修饰的内容是可以改变的。
  char buf[512];      
  strcpy(buf, path);
  char *p = buf + strlen(buf);
  if (*(p - 1) != '/')
  {
    *p++ = '/';
  }


  // char *p = path + strlen(path) - 1;
  // if (*p != '/')
  // {
  //   *++p = '/';
  // }
  // p++;

  //int find = 0;
  while (readdir(fd, &st))
  {
    strcpy(p, st.name);
    if (strcmp(p, ".") == 0 || strcmp(p, "..") == 0)
    {
      continue;
    }

    if (memcmp(p, last_word, strlen(last_word)) == 0)
    {
      printf("%s", p + strlen(last_word));
      added_length = strlen(p) - strlen(last_word);
      strcpy(last_word, p);
      //find = 1;
      break;
    }

  }
  // if(!find)
  // {
  //   printf("\n");
  // }
  close(fd);

  return added_length;
}
```

find_last_word：

```c
char *find_last_word(char *buf) {
  char *word = buf + strlen(buf) - 1;
  while(is_valid_identifier_char(*word) && word != buf-1) {
    word--;
  }
  word++;
  return word;
}

int is_valid_identifier_char(char c) { // used in tab completion
  return  (c >= 'A' && c <= 'Z') ||
          (c >= 'a' && c <= 'z') ||
          (c >= '0' && c <= '9') ||
          c == '_' || c == '.' || c == '-';
}
```



做完这些以后按下tab就可以补全了，但是目前还存在bug，当按下tab后，此时你无法使用backspace键删除字符，比如cd a，然后tab补全后为cd aaa，但是此时按下backspace无法删除a，你只能要么继续输入，要么就直接回车。

原因是因为在consoleintr中，添加了cons.w = cons.e。

```c
if(c == '\t')		//特殊处理
      {
        cons.buf[cons.e++ % INPUT_BUF] = c;
        cons.w = cons.e;
        tab_flag = 1;
        wakeup(&cons.r);
        break;
      }
```

这就不得不提到结构体cons

```c
struct {
  struct spinlock lock;
  
  // input
#define INPUT_BUF 128
  char buf[INPUT_BUF];
  uint r;  // Read index
  uint w;  // Write index
  uint e;  // Edit index
} cons;
```

xv6中为了让用户能够对自己输入的内容进行修改，定义了3个索引，r, w, e。e表示编辑索引，每当用户输入一个字符，编辑索引就会+1，但是w不变，直到输入为'\n'等时，才会使得w = e，这时候将唤醒consoleread进行读操作。

```c
if(c == '\n' || c == C('D') || cons.e == cons.r+INPUT_BUF){
        // wake up consoleread() if a whole line (or end-of-file)
        // has arrived.
        cons.w = cons.e;
        wakeup(&cons.r);
      }

consoleread中:

while(cons.r == cons.w ){
      if(myproc()->killed){
        release(&cons.lock);
        return -1;
      }
```

当w和e不相等时，按下backspace才有用。

```c
case C('H'): // Backspace
case '\x7f':
   if(cons.e != cons.w){
      cons.e--;
      consputc(BACKSPACE);
    }
```

这就是为什么对'\t'特殊处理后无法正常删除的原因。目前还没有解决这个问题，但基本的tab补全已实现，只是无法正常删除，对于使用影响不是很大。