# 静态多级反馈队列

在系统中设置每个优先级对应的等待队列，进程初始化时进入某个队列。调度时，首先从最高优先级队列中找进程，只有当更高优先级队列为空时，才会调度某个低优先级队列中的进程。 同一个优先级的进程按FCFS调度。

多级反馈队列也有不同的实现。静态实现时，进程优先级不在运行过程中自动改变；动态实现时，可以让进程每运行完k个时间片，就下降到低一级的等待队列，一段时间后，再将进程提高到最高优先级，这样可减少进程饥饿的发生。目前我们实现的是静态多级反馈队列。



## 数据结构

- proc结构体定义

  ```c
  struct proc {
    struct spinlock lock;
    ......
    ......
  
    uint ctime;                  // 创建时间
    uint rutime;                 //处于RUNNING状态的时间
    uint retime;                 //处于READY状态的时间
    uint sltime;                 //处于SLEEPING状态的时间
  
    int priority;				   //进程的优先级[1-4]
  };
  ```

- 为每个优先级定义对应的等待队列

  ```c
  typedef struct queue_t
  {
      int size;
      struct proc* proc[NPROC];
      int priority;
  }queue_t;
  
  
  在proc.c中定义4个对应的等待队列
  struct queue_t pqueue_1;
  struct queue_t pqueue_2;
  struct queue_t pqueue_3;
  struct queue_t pqueue_4;
  ```



## 算法

- 内核开始时，要初始化4条优先级队列，定义一个初始化函数来完成：

```c
void
pqueue_init(struct queue_t *queue, int priority)
{
   queue->priority=priority;     //为方便输出debug，加入变量priority
   queue->size=0;
   for(int i=0; i<NPROC; i++)    //初始时指针为0
      queue->proc[i]=0;
}

void
pinit(void)
{				
  pqueue_init(&pqueue_1,1);
  pqueue_init(&pqueue_2,2);
  pqueue_init(&pqueue_3,3);
  pqueue_init(&pqueue_4,4);
}
```



- 入队列：

由于进程的优先级在运行过程中可能动态变化，所以可能要将一个进程从某个队列删除后添加到另一个队列中，这时，为了保持队列从到达时间早到晚排序，要遍历队列，找到该进程的插入位置。从最后（q->size-1)往前扫描，如果当前指针进程的到达时间ctime比新进程p->ctime大（晚），就把它往后挪。直到指向一个进程，到达时间<=p->ctime，然后把p插入到它的后面。最后，要将q->size增加一。

```c
void 
enqueue(struct queue_t *q, struct proc *p)
{
    int pos;
    pos = q->size - 1;
    while(pos >= 0 && q->proc[pos]->ctime > p->ctime){   //从后往前扫描
        q->proc[pos+1] = q->proc[pos];    //如果目前进程比新插入进程到达晚，就把它向后挪一个位置
		pos--;
    }
    // 目前pos指向的进程比插入的进程到达时间早或相同。
    q->proc[pos+1] = p;   
    q->size ++;   
}
```

具体步骤如下：

1. 初始化位置指针：pos 被初始化为 q->size - 1，即队列中最后一个元素的位置。
2. 从后往前扫描：通过 while 循环，从队列的末尾开始向前扫描，比较每个进程的到达时间 ctime。
3. 如果当前扫描到的进程的到达时间 q->proc[pos]->ctime 大于新插入进程的到达时间 p->ctime，则将当前进程向后移动一个位置（q->proc[pos+1] = q->proc[pos]），并将位置指针 pos 向前移动（pos--）。
4. 插入新进程：当找到一个到达时间小于或等于新插入进程的进程时，停止扫描，并将新进程插入到该位置的后面（q->proc[pos+1] = p）。
5. 更新队列大小：最后，增加队列的大小（q->size++）。



- 删除队列

从前往后遍历该等待队列，找到要删除的进程，如果能找到，则将从该位置开始后面的进程都往前挪一个位置，最后一个位置变为0，最后把size减1.

```c
void
removeq(struct queue_t* q, struct proc *p)
{
   int pos;
   int found = 0;

   pos = 0;
   while(pos < q->size){
      if(q->proc[pos]==p){   //从前往后，找到该进程
          found=1;
          break;
      }
      else{
        pos++;
      }
   }
   if(found){   //若找到，此时pos指向要删除的进程
      while(pos < q->size-1){    //从 pos 到倒数第二个进程位置
            q->proc[pos] = q->proc[pos+1];  //将所有进程往前挪一个位置
            pos++;
      }
      q->proc[q->size-1] = 0;
      q->size --;
    }
}
```



- 出队列

查找等待队列中第一个RUNNABLE的进程，只要从前往后遍历，找到第一个RUNNABLE的进程就返回即可。如果未找到，则返回0.

```c
struct proc*
headq(struct queue_t* q)
{
    struct proc* p = 0;
    int pos;

    for(pos = 0; pos < q->size; pos++){
      if(q->proc[pos]->state == RUNNABLE){    //找到第一个RUNNABLE进程，注意等待队列是从到达时间小到大排序的
        p = q->proc[pos];
        break;
      }
    }

    return p;
}
```



- 更新进程运行时间

update()函数添加在timer_tick中

```c
void 
update()
{
  struct proc *p;

  for (p = proc; p < &proc[NPROC]; p++)
  {
    acquire(&p->lock); 
    switch (p->state)
    {              // 遍历整个表，判断进程状态
    case RUNNING:  // 运行中
      p->rutime++; // 则运行时间增加一个时间片
      break;
    case SLEEPING: // 休眠中
      p->sltime++; // 休眠时间增加一个时间片
      break;
    case RUNNABLE: // 等待中
      p->retime++; // 等待时间增加一个时间片
      break;
    default:; // default，有ZOMBIE\EMBRYO\UNUSED，这些状态我们不关心
    }
    release(&p->lock); // 释放锁
  }
  
}


void timer_tick() {
    acquire(&tickslock);
    ticks++;
    update();
    wakeup(&ticks);
    release(&tickslock);
    set_next_timeout();
}
```



接下来要就是要实现进程在执行过程中状态或优先级变化时相应的等待队列的操作。由于等待队列代价比较高，所以尽量在确定进程已经正确分配好各种资源，才将它加入到初始优先级队列中去。 在userinit中，设置默认优先级为2：

```c
void
userinit(void)
{
  ....
  p->priority = 2;
  enqueue(&pqueue_2, p);
  ....
}
```



在fork()中，让子进程继承父进程的优先级（默认为2），然后进入相应的等待队列:

```c
int
fork(void)
{
  ......
  np->priority = p->priority;

  switch (np->priority)
  {
  case 1:
    enqueue(&pqueue_1, np);
    break;
  case 2:
    enqueue(&pqueue_2, np);
    break;
  case 3:
    enqueue(&pqueue_3, np);
    break;
  case 4:
    enqueue(&pqueue_4, np);
    break;  
  default:
    panic("priority should be 1-4\n");
  }
  .....
}

```



在进程终止，被父进程回收资源时，要将它们从相应的等待队列中删除:

```c
static void
freeproc(struct proc *p)
{
  .....
  switch (p->priority)
  { // 删除等待队列中的进程
  case 1:
    removeq(&pqueue_1, p);
    break;
  case 2:
    removeq(&pqueue_2, p);
    break;
  case 3:
    removeq(&pqueue_3, p);
    break;
  case 4:
    removeq(&pqueue_4, p);
    break;
  default:;
  }
}
```



调度器：

从最高优先级队列开始找起，如果某个队列里面有RUNNABLE的进程，就运行它，否则就往下一个优先级队列继续寻找：

```c
void
scheduler(void)
{
  struct proc *p;
  struct cpu *c = mycpu();
  extern pagetable_t kernel_pagetable;

  c->proc = 0;
  for (;;)
  {
    intr_on();
    acquire(&pqueue_lock);
    //寻找优先进程
    p = headq(&pqueue_1);
    if (p == 0)
    {
      p = headq(&pqueue_2);
    }
    if (p == 0)
    {
      p = headq(&pqueue_3);
    }
    if (p == 0)
    {
      p = headq(&pqueue_4);
    }
    release(&pqueue_lock);
    if (p)    //添加last_proc是否处于RUNABLE状态
    { 
      //printf("%d is being scheduled on cpu: %d, its priority is %d.\n", p->pid, cpuid(), p->priority);
      
      acquire(&p->lock);

      //这里判断当前取出来的这个进程是否为RUNNABLE状态，否则两个CPU同时运行一个进程将会出现问题
      if(p->state != RUNNABLE) {		
        release(&p->lock);
        continue;
      }

      p->state = RUNNING;
      c->proc = p;
      w_satp(MAKE_SATP(p->kpagetable));
      sfence_vma();
      swtch(&c->context, &p->context);
      w_satp(MAKE_SATP(kernel_pagetable));
      sfence_vma();

      c->proc = 0;
      release(&p->lock);
    }

  }
}
```



## 系统调用

- waitSH：和wait差不多，只是在这里继续更新进程状态

```c
int waitSH(int* rutime, int* retime, int* sltime)
{
  struct proc *np;
  int havekids, pid;
  struct proc *p = myproc();

  // hold p->lock for the whole time to avoid lost
  // wakeups from a child's exit().
  acquire(&p->lock);

  for(;;){
    // Scan through table looking for exited children.
    havekids = 0;
    for(np = proc; np < &proc[NPROC]; np++){
      // this code uses np->parent without holding np->lock.
      // acquiring the lock first would cause a deadlock,
      // since np might be an ancestor, and we already hold p->lock.
      if(np->parent == p){
        // np->parent can't change between the check and the acquire()
        // because only the parent changes it, and we're the parent.
        acquire(&np->lock);
        havekids = 1;
        if(np->state == ZOMBIE){
          // Found one.
          pid = np->pid;
          *rutime = np->rutime;		//不同于wait
          *retime = np->retime;
          *sltime = np->sltime;
          freeproc(np);
          release(&np->lock);
          release(&p->lock);
          return pid;
        }
        release(&np->lock);
      }
    }

    // No point waiting if we don't have any children.
    if(!havekids || p->killed){
      release(&p->lock);
      return -1;
    }
    
    // Wait for a child to exit.
    sleep(p, &p->lock);  //DOC: wait-sleep
  }
}

```

- set_priority: 可手动设置进程的优先级

```c
int set_priority(int pid, int priority)
{
  struct proc *p = 0;
  
  //遍历进程表，找到匹配的进程
  for (p = proc; p < &proc[NPROC]; p++)
  {
    acquire(&p->lock);

    if (p->pid == pid)		//找到匹配的进程
    {

      if (p->priority != priority)
      {
        //根据当前优先级从对应队列中移除进程
        switch (p->priority)
        {
        case 1:
          removeq(&pqueue_1, p);
          break;
        case 2:
          removeq(&pqueue_2, p);
          break;
        case 3:
          removeq(&pqueue_3, p);
          break;
        case 4:
          removeq(&pqueue_4, p);
          break;
        default:;
        }
        //根据新的优先级将进程加入对应队列
        switch (priority)
        {
        case 1:
          enqueue(&pqueue_1, p);
          break;
        case 2:
          enqueue(&pqueue_2, p);
          break;
        case 3:
          enqueue(&pqueue_3, p);
          break;
        case 4:
          enqueue(&pqueue_4, p);
          break;
        default:;
        }
      }

      p->priority = priority;	//更新进程的优先级

      release(&p->lock);
      return 0;
    }
    release(&p->lock);
  }

  return -1;
}
```



编写用户程序进行测试：

change_priority.c:

```c
#include "kernel/include/types.h"
#include "kernel/include/stat.h"
#include "xv6-user/user.h"

int main(int argc, char* argv[])
{
    if(argc!=2){
        printf("usage: priority_sta <fork number> \n");
	exit(0);
    }
    int forknumber=atoi(argv[1]);
    int status;

    for(int i=0;i<forknumber;i++)
    {
        printf("fork %d\n", i);
        status=fork();
        if(status)
		{
            printf("%d start to set priority\n",status);
	    	set_priority(status, (i+3)%4+1);   //在父进程中，将每个子进程的优先级设置为(i+3)%4+1.
            printf("%d set priority done\n", status);
        }
        if(status==0){
	    	for(int count=0;count<40; count++){
				printf(" ");
		    	for(int k=0;k<1000000;k++){
					int result=0;
					for( int j=0; j<5000; j++){
			    		result+=j*j;
                        result-=j*j;
                        result= result/j;
                        result=result*j;
                        result = result+result/j;
					}
		   		}
            }
            exit(0);
        }
    }
    int rutime;
    int retime;
    int sltime;
    int pid;
    int sum=0;
    while((pid = waitsch(&rutime,&retime,&sltime))!=-1){
        sum+= rutime+retime+sltime;
    }
    printf(" average turn-around time of %d process is %d\n", forknumber, sum/forknumber);
    exit(0);
}
```

结果：

![image-20241224145111217](./picture/静态多级反馈队列.png)

可以看到虽然进程创建顺序为6789，但是修改其优先级后执行顺序为7869。