# 一、进程管理原理

## **第一个用户进程的创建**

**用户进程初始化：**userinit函数通过`allocproc`分配进程并设置页表，通过`uvminit`函数为用户进程分配物理页，加载用户程序代码。设置`trapframe->epc`为用户程序的入口地址（虚拟地址），通过页表将虚拟地址映射到物理内存。

**地址映射与MMU：**通过MMU，将用户虚拟地址（如`trapframe->epc`）转换为物理地址，确保能够正确访问用户程序。

## 进程调度与上下文切换

**调度器`scheduler`**：通过检查进程队列中的`running`状态进程，决定哪个进程应该运行。通过`swtch`实现把当前上下文切换到下一个进程的上下文。

**进程上下文初始化**：在`allocproc`中，设置新进程的`context.ra`为`forkret`，使得上下文切换后能够正确运行到内核指定的位置。

## 进入用户态执行用户程序

**内核到用户态的过渡**：通过`usertrapret`进行进入用户态前的准备，包括设置中断、异常入口，以及页表和用户态寄存器的初始化。通过`satp`寄存器加载页表地址，再通过`sfence.vma`清空TLB。

**用户程序执行**：最终执行`sret`指令切换到用户态，CPU将从`trapframe->epc`指向的用户程序虚拟地址开始执行，通过MMU映射到实际的物理地址，读取用户代码并运行。

## 用户使用过程

用户进行进程管理的核心是利用进程的创建、调度、执行及销毁等功能。在`xv6`中，进程管理主要通过以下步骤实现：

1. **进程创建**：通过系统调用`fork`、等创建新进程或线程。
2. **进程切换**：通过调度器在就绪的进程之间切换。
3. **进程执行**：通过`exec`执行用户程序。
4. **进程回收**：通过wait等待子进程退出并清理其资源。

# 二、扩展的功能

为了给进程管理添加一些经常使用的功能，我们在新增了`sys_execve`、`sys_getppid`、`sys_clone`、`sys_waitpid`系统调用

## **1.execve 系统调用**

**功能：**`execve`系统调用用于指定自定义环境变量加载并执行新的用户程序，替换当前进程的内存镜像，同时保留进程ID。

### **实现流程**

1. 参数解析：从用户空间获取程序路径`path`、参数列表`argv`和环境变量`envp`的地址。将参数和环境变量从用户空间拷贝到内核缓冲区。
2. 调用核心执行函数：调用内核`exec`函数加载新程序。
3. 资源清理：释放为`argv`和`envp`分配的内存。
4. 返回结果：若成功，返回执行结果；否则返回-1。

### **代码分析**

```
uint64 sys_execve(void) {
    // 1. 获取用户参数
    if (argstr(0, path, FAT32_MAX_PATH) < 0 || argaddr(1, &uargv) < 0 || argaddr(2, &uenvp) < 0) {
        return -1;
    }
    ...
    // 2. 加载参数列表和环境变量
    for (i = 0; i < MAXARG; i++) {
        if (fetchaddr(uargv + sizeof(uint64) * i, &uarg) < 0 || fetchstr(uarg, argv[i], PGSIZE) < 0) {
            goto bad;
        }
    }
    // 3. 调用核心 exec 函数
    int ret = exec(path, argv);
    ...
    // 4. 清理内存
    for (i = 0; i < MAXARG && argv[i] != 0; i++) {
        kfree(argv[i]);
    }
    return ret;
}
```



## **2.getppid 系统调用**

**功能：**`getppid`系统调用用于获取当前进程的父进程ID。

### **实现流程**

1. 获取当前进程的`proc`结构。
2. 加锁以访问`parent`字段。
3. 获取父进程的`pid`并返回。

### **代码分析**

比较简单，直接获取父进程的ID返回即可

```
uint64 sys_getppid(void) {
    struct proc *p = myproc();
    acquire(&p->lock);  // 加锁保护
    uint64 ppid = p->parent->pid;  // 获取父进程 PID
    release(&p->lock);
    return ppid;
}
```

**关键点：**加锁以保证对`parent`字段的访问线程安全。

测试程序：testgetppid

## **3.clone 系统调用**

**功能：**`clone`系统调用用于创建一个共享地址空间的线程，与`fork`不同，线程的地址空间是与父进程共享的。

### **实现流程**

1. 参数解析：从用户空间获取线程栈地址`stack`、入口函数地址`func`以及函数参数`arg`。检查栈地址是否对齐。

2. 线程创建：调用`allocproc`分配新的`proc`结构。共享父进程的地址空间、文件描述符表等资源。

3. 设置线程上下文：设置线程的栈指针`sp`、入口地址`epc`和函数参数。将线程标记为`RUNNABLE`。

4. **返回线程ID**。

    

**对应代码**

```
uint64 sys_clone(void) {
    if (argaddr(0, &stack) < 0 || argaddr(1, &func) < 0 || argaddr(2, &arg)) {
        return -1;
    }
    return clone(stack, func, arg);
}

int clone(uint64 stack, uint64 func, uint64 arg) {
    struct proc *np;
    struct proc *curproc = myproc();
    if (stack % 256 != 0) {  // 栈地址对齐检查
        return -1;
    }
    if ((np = allocproc()) == 0) {
        return -1;
    }
    // 共享地址空间和文件表
    np->pagetable = curproc->pagetable;
    np->sz = curproc->sz;
    ...
    np->trapframe->epc = func;  // 设置入口函数
    np->trapframe->sp = stack;  // 设置栈指针
    np->trapframe->a0 = arg;    // 设置函数参数
    release(&np->lock);
    return np->pid;  // 返回线程 ID
}
```



实现的过程中还需要对进程结构体进行改造，添加两个线程字段，标记线程以及后续扩展

```
// proc.h
struct proc {
  // 原有字段
  uint64 sz;               // 进程地址空间大小
  pagetable_t pagetable;   // 页表
  struct trapframe *tf;    // 中断帧
  int killed;              // 是否被杀死
  struct proc *parent;     // 父进程

  // 新增字段
  int is_thread;           // 标志是否是线程（1 为线程，0 为进程）
  void *user_stack;        // 线程的用户栈指针
};
```

同时新增 `thread_exit` 和 `thread_join`两个系统调用实现线程的等待和退出

 `thread_join`：等待某个线程退出并回收其资源。

```
int
thread_join(int pid)
{
  struct proc *p;
  for (;;) {
    // 遍历进程表，找到目标线程
    for (p = proc; p < &proc[NPROC]; p++) {
      acquire(&p->lock);
      if (p->pid == pid && p->is_thread) {
        // 如果线程已经退出
        if (p->state == ZOMBIE) {
          // 释放线程的资源
          freeproc(p);
          release(&p->lock);
          return 0; // 成功回收线程
        }
        // 如果线程未退出，解锁并继续等待
        // 如果目标线程未退出，进入休眠等待
        sleep(myproc(), &p->lock);// 当前线程睡眠，等待被唤醒
        break;
      }
      release(&p->lock); // 解锁，继续检查下一个线程
    }
  }
}
```

 `thread_exit` ：线程退出时需要释放自身的资源，但不破坏共享资源（如虚拟内存、文件表）。

```
void
thread_exit(void)
{
  struct proc *p = myproc();

  // 标记为 ZOMBIE
  acquire(&p->lock);
  p->state = ZOMBIE;
  // 如果线程分配了独立的用户栈，释放它
  if (p->is_thread && p->user_stack) {
    kfree(p->user_stack);
    p->user_stack = 0;
  }
  // 唤醒等待线程的父进程或主线程
  wakeup(p->parent);
  release(&p->lock);
  // 调度其他进程或线程
  sched();
  return;
}
```

测试程序：testclone，用户程序中无法使用函数地址，子线程没有真正运行起来

## **4.waitpid 系统调用**

**功能：**`waitpid`系统调用用于等待指定子进程退出，并回收其资源。与wait不同的是，可以等待指定的子进程退出。

### **实现流程**

1. 参数解析：获取子进程ID、退出状态指针以及选项。
2. 遍历进程表：遍历`proc`表，查找符合条件的子进程。
3. 等待退出：若子进程已退出，拷贝退出状态并清理资源。若未退出，进入休眠等待。
4. **返回退出的子进程ID**。

### **代码分析**

sys_waitpid是系统调用所对应函数，在其内部首先获取传入的参数，包括等待的子进程id，子进程返回后的状态status，可选参数options(可扩展实现立即返回等待，阻塞等待等功能)。然后调用内核中的函数waitpid执行具体逻辑。

```
uint64
sys_waitpid(void)
{
  int pid,options;
  uint64 status;
  if(argint(0, &pid) < 0)
    return -1;
  if(argaddr(1, &status) < 0)
    return -1;
  if(argint(2,&options) < 0)
    return -1;
  return waitpid(pid,status,options);
}

```

waitpid：内核函数，遍历进程数组查找对应pid的进程，若该进程未退出，进入等待状态；如果已经退出，返回该进程退出时的状态

```
int waitpid(int pid, uint64 addr, int options) {
    struct proc *p = myproc();
    acquire(&p->lock);
    for (;;) {
        havekids = 0;
        for (np = proc; np < &proc[NPROC]; np++) {
            if (np->parent == p && (pid == -1 || np->pid == pid)) {
                acquire(&np->lock);
                havekids = 1;
                if (np->state == ZOMBIE) {  // 子进程退出
                    npid = np->pid;
                    int exit_status = np->xstate;
                    if (addr != 0 && copyout2(addr, (char *)&exit_status, sizeof(exit_status)) < 0) {
                        release(&np->lock);
                        release(&p->lock);
                        return -1;
                    }
                    freeproc(np);  // 清理资源
                    release(&np->lock);
                    release(&p->lock);
                    return npid;  // 返回子进程 ID
                }
                release(&np->lock);
            }
        }
        if (!havekids || p->killed) {  // 无子进程或父进程被杀
            release(&p->lock);
            return -1;
        }
        sleep(p, &p->lock);
    }
}
```

测试程序：testwaitpid

# 三、总结

1. **execve**：用于携带指定的环境变量执行新程序，涉及用户参数拷贝和程序加载。
2. **getppid**：快速获取父进程ID，简洁高效。
3. **clone**：用于线程创建，实现地址空间共享和轻量化。
4. **waitpid**：等待特定子进程的退出，完成状态传递和资源回收。