
#include "include/types.h"
#include "include/param.h"
#include "include/memlayout.h"
#include "include/riscv.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/intr.h"
#include "include/kalloc.h"
#include "include/printf.h"
#include "include/string.h"
#include "include/fat32.h"
#include "include/file.h"
#include "include/trap.h"
#include "include/vm.h"
#include "include/timer.h"


struct cpu cpus[NCPU];

struct proc proc[NPROC];

struct proc *initproc;

struct queue_t pqueue_1;
struct queue_t pqueue_2;
struct queue_t pqueue_3;
struct queue_t pqueue_4;

int nextpid = 1;
struct spinlock pid_lock;
struct spinlock pqueue_lock;

extern void forkret(void);
extern void swtch(struct context*, struct context*);
static void wakeup1(struct proc *chan);
static void freeproc(struct proc *p);

extern char trampoline[]; // trampoline.S

void reg_info(void) {
  printf("register info: {\n");
  printf("sstatus: %p\n", r_sstatus());
  printf("sip: %p\n", r_sip());
  printf("sie: %p\n", r_sie());
  printf("sepc: %p\n", r_sepc());
  printf("stvec: %p\n", r_stvec());
  printf("satp: %p\n", r_satp());
  printf("scause: %p\n", r_scause());
  printf("stval: %p\n", r_stval());
  printf("sp: %p\n", r_sp());
  printf("tp: %p\n", r_tp());
  printf("ra: %p\n", r_ra());
  printf("}\n");
}

// initialize the proc table at boot time.
void
procinit(void)
{
  struct proc *p;
  
  initlock(&pid_lock, "nextpid");
  initlock(&pqueue_lock, "pqueue");
  for(p = proc; p < &proc[NPROC]; p++) {
      initlock(&p->lock, "proc");

      // Allocate a page for the process's kernel stack.
      // Map it high in memory, followed by an invalid
      // guard page.
      // char *pa = kalloc();
      // // printf("[procinit]kernel stack: %p\n", (uint64)pa);
      // if(pa == 0)
      //   panic("kalloc");
      // uint64 va = KSTACK((int) (p - proc));
      // // printf("[procinit]kvmmap va %p to pa %p\n", va, (uint64)pa);
      // kvmmap(va, (uint64)pa, PGSIZE, PTE_R | PTE_W);
      // p->kstack = va;
  }
  //kvminithart();

  memset(cpus, 0, sizeof(cpus));
  #ifdef DEBUG
  printf("procinit\n");
  #endif
}

// Must be called with interrupts disabled,
// to prevent race with process being moved
// to a different CPU.
int
cpuid()
{
  int id = r_tp();
  return id;
}

// Return this CPU's cpu struct.
// Interrupts must be disabled.
struct cpu*
mycpu(void) {
  int id = cpuid();
  struct cpu *c = &cpus[id];
  
  return c;
}

// Return the current struct proc *, or zero if none.
struct proc*
myproc(void) {
  push_off();
  struct cpu *c = mycpu();
  struct proc *p = c->proc;
  pop_off();
  return p;
}

int
allocpid() {
  int pid;
  
  acquire(&pid_lock);
  pid = nextpid;
  nextpid = nextpid + 1;
  release(&pid_lock);

  return pid;
}

// Look in the process table for an UNUSED proc.
// If found, initialize state required to run in the kernel,
// and return with p->lock held.
// If there are no free procs, or a memory allocation fails, return 0.
static struct proc*
allocproc(void)
{
  struct proc *p;

  for(p = proc; p < &proc[NPROC]; p++) {
    acquire(&p->lock);
    if(p->state == UNUSED) {
      goto found;
    } else {
      release(&p->lock);
    }
  }
  return NULL;

found:
  p->pid = allocpid();
  p->ctime = ticks;
  p->rutime = 0;
  p->retime = 0;
  p->sltime = 0;
  p->gid = 0;
  p->uid = 0;
  


  // Allocate a trapframe page.
  if((p->trapframe = (struct trapframe *)kalloc()) == NULL){
    release(&p->lock);
    p->ctime = 0;
    return NULL;
  }

  // An empty user page table.
  // And an identical kernel page table for this proc.
  if ((p->pagetable = proc_pagetable(p)) == NULL ||
      (p->kpagetable = proc_kpagetable()) == NULL) {
    freeproc(p);
    release(&p->lock);
    p->ctime = 0;
    return NULL;
  }
  for(int i = 0; i < MMAPNUM; ++i){
    p->mmap_pool[i].used = 0;
  }
  p->kstack = VKSTACK;

  // Set up new context to start executing at forkret,
  // which returns to user space.
  memset(&p->context, 0, sizeof(p->context));
  p->context.ra = (uint64)forkret;
  p->context.sp = p->kstack + PGSIZE;

  return p;
}

// free a proc structure and the data hanging from it,
// including user pages.
// p->lock must be held.
static void
freeproc(struct proc *p)
{
  if(p->trapframe)
    kfree((void*)p->trapframe);
  p->trapframe = 0;
  if (p->kpagetable) {
    kvmfree(p->kpagetable, 1);
  }
  p->kpagetable = 0;
  if(p->pagetable)
    proc_freepagetable(p->pagetable, p->sz);
  p->pagetable = 0;
  p->sz = 0;
  p->pid = 0;
  p->parent = 0;
  p->name[0] = 0;
  p->chan = 0;
  p->killed = 0;
  p->xstate = 0;
  p->state = UNUSED;

  p->ctime = 0;
  p->rutime = 0;
  p->retime = 0;
  p->sltime = 0;

  switch (p->priority)
  { // 删除等待队列中的进程
  case 1:
    removeq(&pqueue_1, p);
    break;
  case 2:
    removeq(&pqueue_2, p);
    break;
  case 3:
    removeq(&pqueue_3, p);
    break;
  case 4:
    removeq(&pqueue_4, p);
    break;
  default:;
  }
}

// Create a user page table for a given process,
// with no user memory, but with trampoline pages.
pagetable_t
proc_pagetable(struct proc *p)
{
  pagetable_t pagetable;

  // An empty page table.
  pagetable = uvmcreate();
  if(pagetable == 0)
    return NULL;

  // map the trampoline code (for system call return)
  // at the highest user virtual address.
  // only the supervisor uses it, on the way
  // to/from user space, so not PTE_U.
  if(mappages(pagetable, TRAMPOLINE, PGSIZE,
              (uint64)trampoline, PTE_R | PTE_X) < 0){
    uvmfree(pagetable, 0);
    return NULL;
  }

  // map the trapframe just below TRAMPOLINE, for trampoline.S.
  if(mappages(pagetable, TRAPFRAME, PGSIZE,
              (uint64)(p->trapframe), PTE_R | PTE_W) < 0){
    vmunmap(pagetable, TRAMPOLINE, 1, 0);
    uvmfree(pagetable, 0);
    return NULL;
  }

  return pagetable;
}

// Free a process's page table, and free the
// physical memory it refers to.
void
proc_freepagetable(pagetable_t pagetable, uint64 sz)
{
  vmunmap(pagetable, TRAMPOLINE, 1, 0);
  vmunmap(pagetable, TRAPFRAME, 1, 0);
  uvmfree(pagetable, sz);
}

// a user program that calls exec("/init")
// od -t xC initcode
uchar initcode[] = {
  0x17, 0x05, 0x00, 0x00, 0x13, 0x05, 0x45, 0x02,
  0x97, 0x05, 0x00, 0x00, 0x93, 0x85, 0x35, 0x02,
  0x93, 0x08, 0x70, 0x00, 0x73, 0x00, 0x00, 0x00,
  0x93, 0x08, 0x20, 0x00, 0x73, 0x00, 0x00, 0x00,
  0xef, 0xf0, 0x9f, 0xff, 0x2f, 0x69, 0x6e, 0x69,
  0x74, 0x00, 0x00, 0x24, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00
};

// uchar printhello[] = {
//     0x13, 0x00, 0x00, 0x00,     // nop
//     0x13, 0x00, 0x00, 0x00,     // nop 
//     0x13, 0x00, 0x00, 0x00,     // nop 
//     // <start>
//     0x17, 0x05, 0x00, 0x00,     // auipc a0, 0x0 
//     0x13, 0x05, 0x05, 0x00,     // mv a0, a0 
//     0x93, 0x08, 0x60, 0x01,     // li a7, 22 
//     0x73, 0x00, 0x00, 0x00,     // ecall 
//     0xef, 0xf0, 0x1f, 0xff,     // jal ra, <start>
//     // <loop>
//     0xef, 0x00, 0x00, 0x00,     // jal ra, <loop>
// };


// void test_proc_init(int proc_num) {
//   if(proc_num > NPROC) panic("test_proc_init\n");
//   struct proc *p;
//   for(int i = 0; i < proc_num; i++) {
//     p = allocproc();
//     uvminit(p->pagetable, (uchar*)printhello, sizeof(printhello));
//     p->sz = PGSIZE;
//     p->trapframe->epc = 0x0;
//     p->trapframe->sp = PGSIZE;
//     safestrcpy(p->name, "test_code", sizeof(p->name));
//     p->state = RUNNABLE;
//     release(&p->lock);
//   }
//   initproc = proc;
//   printf("[test_proc]test_proc init done\n");
// }

// Set up first user process.
void
userinit(void)
{
  pinit();
  struct proc *p;

  p = allocproc();
  initproc = p;
  
  // allocate one user page and copy init's instructions
  // and data into it.
  uvminit(p->pagetable , p->kpagetable, initcode, sizeof(initcode));
  p->sz = PGSIZE;

  // prepare for the very first "return" from kernel to user.
  p->trapframe->epc = 0x0;      // user program counter
  p->trapframe->sp = PGSIZE;  // user stack pointer

  safestrcpy(p->name, "initcode", sizeof(p->name));

  p->state = RUNNABLE;

  p->tmask = 0;

  p->priority = 2;
  enqueue(&pqueue_2, p);


  release(&p->lock);
  #ifdef DEBUG
  printf("userinit\n");
  #endif
}

// Grow or shrink user memory by n bytes.
// Return 0 on success, -1 on failure.
int
growproc(int n)
{
  uint sz;
  struct proc *p = myproc();

  sz = p->sz;
  if(n > 0){
    if((sz = uvmalloc(p->pagetable, p->kpagetable, sz, sz + n)) == 0) {
      return -1;
    }
  } else if(n < 0){
    sz = uvmdealloc(p->pagetable, p->kpagetable, sz, sz + n);
  }
  p->sz = sz;
  return 0;
}

// Create a new process, copying the parent.
// Sets up child kernel stack to return as if from fork() system call.
int
fork(void)
{
  int i, pid;
  struct proc *np;
  struct proc *p = myproc();

  // Allocate process.
  if((np = allocproc()) == NULL){
    return -1;
  }

  // Copy user memory from parent to child.
  if(uvmcopy(p->pagetable, np->pagetable, np->kpagetable, p->sz) < 0){
    freeproc(np);
    release(&np->lock);
    return -1;
  }
  np->sz = p->sz;

  np->parent = p;

  // copy tracing mask from parent.
  np->tmask = p->tmask;

  // copy saved user registers.
  *(np->trapframe) = *(p->trapframe);

  // Cause fork to return 0 in the child.
  np->trapframe->a0 = 0;

  // increment reference counts on open file descriptors.
  for(i = 0; i < NOFILE; i++)
    if(p->ofile[i])
      np->ofile[i] = filedup(p->ofile[i]);
  np->cwd = edup(p->cwd);

  safestrcpy(np->name, p->name, sizeof(p->name));

  pid = np->pid;

  np->state = RUNNABLE;

  np->priority = p->priority;

  switch (np->priority)
  {
  case 1:
    enqueue(&pqueue_1, np);
    break;
  case 2:
    enqueue(&pqueue_2, np);
    break;
  case 3:
    enqueue(&pqueue_3, np);
    break;
  case 4:
    enqueue(&pqueue_4, np);
    break;  
  default:
    panic("priority should be 1-4\n");
  }

  release(&np->lock);

  return pid;
}

int
clone(uint64 stack, uint64 func,uint64 arg)
{
  struct proc *np;
  struct proc *curproc = myproc();
  // 打印调用 clone 的线程和参数信息
  //printf("clone: called by pid %d with stack=0x%d, func=0x%d, arg=0x%d\n",curproc->pid, stack, func, arg);
  if (stack % 256 != 0) {
    printf("stack = %d\n",stack);
    printf("Error: stack address not aligned\n");
    return -1;
  }
  // 打印传入的栈地址
  //printf("clone: received stack address = %d\n", stack);
  // 创建新线程
  if ((np = allocproc()) == 0)
    return -1;

  // 共享地址空间
  np->pagetable = curproc->pagetable;
  np->kpagetable =curproc->kpagetable;
  np->sz = curproc->sz;

  // 设置父子关系
  np->parent = curproc;
  safestrcpy(np->name, curproc->name, sizeof(curproc->name));
  // 复制 tmask
  np->tmask = curproc->tmask;  

  // 设置线程的栈指针
  *(np->trapframe) = *(curproc->trapframe);
  np->trapframe->epc = func;
  np->trapframe->sp = stack;
  np->trapframe->a0 = arg;

  // 共享文件表
  for (int i = 0; i < NOFILE; i++) {
    if (curproc->ofile[i])
      np->ofile[i] = curproc->ofile[i];
  }
  np->cwd = edup(curproc->cwd);

  // 设置线程相关信息
  np->is_thread = 1;
  np->user_stack = (void *)stack;

  // 设置线程状态为 RUNNABLE
  //np->state = RUNNABLE;
  release(&np->lock);
  printf("clone: thread created with pid %d\n", np->pid);
  return np->pid;// 返回线程 ID
}

void
thread_exit(void)
{
  struct proc *p = myproc();

  // 标记为 ZOMBIE
  acquire(&p->lock);
  //printf("thread_exit acquire p->lock\n");
  p->state = ZOMBIE;
    
  

  // 如果线程分配了独立的用户栈，释放它
  if (p->is_thread && p->user_stack) {
    kfree(p->user_stack);
    p->user_stack = 0;
  }
  //printf("thread_exit release p->lock\n");
  // 唤醒等待线程的父进程或主线程
  wakeup(p->parent);
  release(&p->lock);
  // 调度其他进程或线程
  sched();

  return;
}

int
thread_join(int pid)
{
  struct proc *p;
  for (;;) {
    // 遍历进程表，找到目标线程
    for (p = proc; p < &proc[NPROC]; p++) {
      //printf("Trying to acquire lock for pid: %d\n", p->pid);
      acquire(&p->lock);
      //printf("Lock acquired for pid: %d\n", p->pid);
      if (p->pid == pid && p->is_thread) {
        //printf("Found target thread with pid: %d\n", p->pid);
        // 如果线程已经退出
        if (p->state == ZOMBIE) {
          //printf("Thread with pid: %d is ZOMBIE, cleaning up\n", p->pid);
          // 释放线程的资源
          freeproc(p);
          release(&p->lock);
          return 0; // 成功回收线程
        }
        // 如果线程未退出，解锁并继续等待
        // 如果目标线程未退出，进入休眠等待
        //printf("Thread with pid: %d not exited, going to sleep\n", p->pid);
        sleep(myproc(), &p->lock);// 当前线程睡眠，等待被唤醒
        //printf("Woke up from sleep for pid: %d\n", p->pid);
        break;
      }
      //printf("5release p->lock\n");
      release(&p->lock); // 解锁，继续检查下一个线程
      //printf("6release p->lock\n");
    }
  }
}

// Pass p's abandoned children to init.
// Caller must hold p->lock.
void
reparent(struct proc *p)
{
  struct proc *pp;

  for(pp = proc; pp < &proc[NPROC]; pp++){
    // this code uses pp->parent without holding pp->lock.
    // acquiring the lock first could cause a deadlock
    // if pp or a child of pp were also in exit()
    // and about to try to lock p.
    if(pp->parent == p){
      // pp->parent can't change between the check and the acquire()
      // because only the parent changes it, and we're the parent.
      acquire(&pp->lock);
      pp->parent = initproc;
      // we should wake up init here, but that would require
      // initproc->lock, which would be a deadlock, since we hold
      // the lock on one of init's children (pp). this is why
      // exit() always wakes init (before acquiring any locks).
      release(&pp->lock);
    }
  }
}

// Exit the current process.  Does not return.
// An exited process remains in the zombie state
// until its parent calls wait().
void
exit(int status)
{
  struct proc *p = myproc();

  if(p == initproc)
    panic("init exiting");

  // Close all open files.
  for(int fd = 0; fd < NOFILE; fd++){
    if(p->ofile[fd]){
      struct file *f = p->ofile[fd];
      fileclose(f);
      p->ofile[fd] = 0;
    }
  }

  eput(p->cwd);
  p->cwd = 0;

  // we might re-parent a child to init. we can't be precise about
  // waking up init, since we can't acquire its lock once we've
  // acquired any other proc lock. so wake up init whether that's
  // necessary or not. init may miss this wakeup, but that seems
  // harmless.
  acquire(&initproc->lock);
  wakeup1(initproc);
  release(&initproc->lock);

  // grab a copy of p->parent, to ensure that we unlock the same
  // parent we locked. in case our parent gives us away to init while
  // we're waiting for the parent lock. we may then race with an
  // exiting parent, but the result will be a harmless spurious wakeup
  // to a dead or wrong process; proc structs are never re-allocated
  // as anything else.
  acquire(&p->lock);
  struct proc *original_parent = p->parent;
  release(&p->lock);
  
  // we need the parent's lock in order to wake it up from wait().
  // the parent-then-child rule says we have to lock it first.
  acquire(&original_parent->lock);

  acquire(&p->lock);

  // Give any children to init.
  reparent(p);

  // Parent might be sleeping in wait().
  wakeup1(original_parent);

  p->xstate = status;
  p->state = ZOMBIE;
  for (int i = 0; i < MMAPNUM; i++)
  {
    if (p->mmap_pool[i].used)
      do_munmap(p->mmap_pool[i].start, p->mmap_pool[i].len);
  }
  release(&original_parent->lock);

  // Jump into the scheduler, never to return.
  sched();
  panic("zombie exit");
}

// Wait for a child process to exit and return its pid.
// Return -1 if this process has no children.
int
wait(uint64 addr)
{
  struct proc *np;
  int havekids, pid;
  struct proc *p = myproc();

  // hold p->lock for the whole time to avoid lost
  // wakeups from a child's exit().
  acquire(&p->lock);

  for(;;){
    // Scan through table looking for exited children.
    havekids = 0;
    for(np = proc; np < &proc[NPROC]; np++){
      // this code uses np->parent without holding np->lock.
      // acquiring the lock first would cause a deadlock,
      // since np might be an ancestor, and we already hold p->lock.
      if(np->parent == p){
        // np->parent can't change between the check and the acquire()
        // because only the parent changes it, and we're the parent.
        acquire(&np->lock);
        havekids = 1;
        if(np->state == ZOMBIE){
          // Found one.
          pid = np->pid;
          if(addr != 0 && copyout2(addr, (char *)&np->xstate, sizeof(np->xstate)) < 0) {
            release(&np->lock);
            release(&p->lock);
            return -1;
          }
          freeproc(np);
          release(&np->lock);
          release(&p->lock);
          return pid;
        }
        release(&np->lock);
      }
    }

    // No point waiting if we don't have any children.
    if(!havekids || p->killed){
      release(&p->lock);
      return -1;
    }
    
    // Wait for a child to exit.
    sleep(p, &p->lock);  //DOC: wait-sleep
  }
}

/* 
  waitpid
  参数pid:要等待的子进程
  addr,存放进程状态的地址，
  options可选参数，扩展更多可选功能，目前没有用到
 */

int 
waitpid(int pid, uint64 addr, int options)
{
  struct proc *np;
  int havekids, npid;
  struct proc *p = myproc();

  acquire(&p->lock); // 获取父进程锁

  for (;;) {
    havekids = 0;
    for (np = proc; np < &proc[NPROC]; np++) { // 遍历进程表
      if (np->parent == p && (pid == -1 || (pid > 0 && np->pid == pid))) {

        acquire(&np->lock); // 获取子进程锁
        havekids = 1;
        if (np->state == ZOMBIE) { // 子进程已退出
          npid = np->pid;
          int exit_status = np->xstate; // 保存退出状态（不移位）
          
          if (addr != 0 && copyout2(addr, (char *)&exit_status, sizeof(exit_status)) < 0) {
            release(&np->lock); // 子进程锁
            release(&p->lock); // 父进程锁
            return -1; // 状态拷贝失败
          }

          freeproc(np); // 清理子进程
          release(&np->lock); // 释放子进程锁
          release(&p->lock); // 释放父进程锁

          return npid; // 返回退出的子进程 PID
        }
        release(&np->lock); // 子进程未退出，释放子进程锁
      }
    }

    // 如果没有子进程或者父进程已被杀，则返回 -1
    if (!havekids || p->killed) {
      release(&p->lock); // 释放父进程锁
      return -1;
    }

    sleep(p, &p->lock);
  }
}


int waitSH(int* rutime, int* retime, int* sltime)
{
  struct proc *np;
  int havekids, pid;
  struct proc *p = myproc();

  // hold p->lock for the whole time to avoid lost
  // wakeups from a child's exit().
  acquire(&p->lock);

  for(;;){
    // Scan through table looking for exited children.
    havekids = 0;
    for(np = proc; np < &proc[NPROC]; np++){
      // this code uses np->parent without holding np->lock.
      // acquiring the lock first would cause a deadlock,
      // since np might be an ancestor, and we already hold p->lock.
      if(np->parent == p){
        // np->parent can't change between the check and the acquire()
        // because only the parent changes it, and we're the parent.
        acquire(&np->lock);
        havekids = 1;
        if(np->state == ZOMBIE){
          // Found one.
          pid = np->pid;
          *rutime = np->rutime;
          *retime = np->retime;
          *sltime = np->sltime;
          freeproc(np);
          release(&np->lock);
          release(&p->lock);
          return pid;
        }
        release(&np->lock);
      }
    }

    // No point waiting if we don't have any children.
    if(!havekids || p->killed){
      release(&p->lock);
      return -1;
    }
    
    // Wait for a child to exit.
    sleep(p, &p->lock);  //DOC: wait-sleep
  }
}


// Per-CPU process scheduler.
// Each CPU calls scheduler() after setting itself up.
// Scheduler never returns.  It loops, doing:
//  - choose a process to run.
//  - swtch to start running that process.
//  - eventually that process transfers control
//    via swtch back to the scheduler.
void
scheduler(void)
{
  struct proc *p;
  struct cpu *c = mycpu();
  extern pagetable_t kernel_pagetable;

  c->proc = 0;
  for (;;)
  {
    intr_on();
    acquire(&pqueue_lock);
    p = headq(&pqueue_1);
    if (p == 0)
    {
      p = headq(&pqueue_2);
    }
    if (p == 0)
    {
      p = headq(&pqueue_3);
    }
    if (p == 0)
    {
      p = headq(&pqueue_4);
    }
    release(&pqueue_lock);
    if (p)    //添加last_proc是否处于RUNABLE状态
    { // 若找得到进程，则运行它，否则暂时释放ptable锁，开启中断，然后再次获取ptable锁，继续从第一优先级队列开始寻找
      //printf("%d is being scheduled on cpu: %d, its priority is %d.\n", p->pid, cpuid(), p->priority);
      
      acquire(&p->lock);

      if(p->state != RUNNABLE) {
        release(&p->lock);
        continue;
      }
      //printf("11111\n");
      p->state = RUNNING;
      c->proc = p;
      w_satp(MAKE_SATP(p->kpagetable));
      //printf("22222\n");
      sfence_vma();
      //printf("33333\n");
      swtch(&c->context, &p->context);
      //printf("44444\n");
      w_satp(MAKE_SATP(kernel_pagetable));
      //printf("55555\n");
      sfence_vma();
      //printf("66666\n");

      // c->proc = p;
      // switchuvm(p);
      // p->state = RUNNING;

      // swtch(&(c->context), &p->context);
      // switchkvm();

      // Process is done running for now.
      // It should have changed its p->state before coming back.
      c->proc = 0;
      release(&p->lock);
    }
    
    
  }

  // c->proc = 0;
  // for(;;){
  //   // Avoid deadlock by ensuring that devices can interrupt.
  //   intr_on();
    
  //   int found = 0;
  //   for(p = proc; p < &proc[NPROC]; p++) {
  //     acquire(&p->lock);
  //     if(p->state == RUNNABLE) {
  //       // Switch to chosen process.  It is the process's job
  //       // to release its lock and then reacquire it
  //       // before jumping back to us.
  //       // printf("[scheduler]found runnable proc with pid: %d\n", p->pid);
  //       p->state = RUNNING;
  //       c->proc = p;
  //       w_satp(MAKE_SATP(p->kpagetable));
  //       sfence_vma();
  //       swtch(&c->context, &p->context);
  //       w_satp(MAKE_SATP(kernel_pagetable));
  //       sfence_vma();
  //       // Process is done running for now.
  //       // It should have changed its p->state before coming back.
  //       c->proc = 0;

  //       found = 1;
  //     }
  //     release(&p->lock);
  //   }
  //   if(found == 0) {
  //     intr_on();
  //     asm volatile("wfi");
  //   }
  // }
}

// Switch to scheduler.  Must hold only p->lock
// and have changed proc->state. Saves and restores
// intena because intena is a property of this
// kernel thread, not this CPU. It should
// be proc->intena and proc->noff, but that would
// break in the few places where a lock is held but
// there's no process.
void
sched(void)
{
  int intena;
  struct proc *p = myproc();

  if(!holding(&p->lock))
    panic("sched p->lock");
  if(mycpu()->noff != 1)
    panic("sched locks");
  if(p->state == RUNNING)
    panic("sched running");
  if(intr_get())
    panic("sched interruptible");

  intena = mycpu()->intena;
  swtch(&p->context, &mycpu()->context);
  mycpu()->intena = intena;
}

// Give up the CPU for one scheduling round.
void
yield(void)
{
  struct proc *p = myproc();
  acquire(&p->lock);
  p->state = RUNNABLE;
  sched();
  release(&p->lock);
}

// A fork child's very first scheduling by scheduler()
// will swtch to forkret.
void
forkret(void)
{
  // printf("run in forkret\n");
  static int first = 1;

  // Still holding p->lock from scheduler.
  release(&myproc()->lock);

  if (first) {
    // File system initialization must be run in the context of a
    // regular process (e.g., because it calls sleep), and thus cannot
    // be run from main().
    // printf("[forkret]first scheduling\n");
    first = 0;
    fat32_init();
    myproc()->cwd = ename("/");
  }

  usertrapret();
}

// Atomically release lock and sleep on chan.
// Reacquires lock when awakened.
void
sleep(void *chan, struct spinlock *lk)
{
  struct proc *p = myproc();
  
  // Must acquire p->lock in order to
  // change p->state and then call sched.
  // Once we hold p->lock, we can be
  // guaranteed that we won't miss any wakeup
  // (wakeup locks p->lock),
  // so it's okay to release lk.
  if(lk != &p->lock){  //DOC: sleeplock0
    acquire(&p->lock);  //DOC: sleeplock1
    release(lk);
  }

  // Go to sleep.
  p->chan = chan;
  p->state = SLEEPING;

  sched();

  // Tidy up.
  p->chan = 0;

  // Reacquire original lock.
  if(lk != &p->lock){
    release(&p->lock);
    acquire(lk);
  }
}

// Wake up all processes sleeping on chan.
// Must be called without any p->lock.
void
wakeup(void *chan)
{
  struct proc *p;

  for(p = proc; p < &proc[NPROC]; p++) {
    acquire(&p->lock);
    if(p->state == SLEEPING && p->chan == chan) {
      p->state = RUNNABLE;
    }
    release(&p->lock);
  }
}

// Wake up p if it is sleeping in wait(); used by exit().
// Caller must hold p->lock.
static void
wakeup1(struct proc *p)
{
  if(!holding(&p->lock))
    panic("wakeup1");
  if(p->chan == p && p->state == SLEEPING) {
    p->state = RUNNABLE;
  }
}

// Kill the process with the given pid.
// The victim won't exit until it tries to return
// to user space (see usertrap() in trap.c).
int
kill(int pid)
{
  struct proc *p;

  for(p = proc; p < &proc[NPROC]; p++){
    acquire(&p->lock);
    if(p->pid == pid){
      p->killed = 1;
      if(p->state == SLEEPING){
        // Wake process from sleep().
        p->state = RUNNABLE;
      }
      release(&p->lock);
      return 0;
    }
    release(&p->lock);
  }
  return -1;
}

// Copy to either a user address, or kernel address,
// depending on usr_dst.
// Returns 0 on success, -1 on error.
int
either_copyout(int user_dst, uint64 dst, void *src, uint64 len)
{
  // struct proc *p = myproc();
  if(user_dst){
    // return copyout(p->pagetable, dst, src, len);
    return copyout2(dst, src, len);
  } else {
    memmove((char *)dst, src, len);
    return 0;
  }
}

// Copy from either a user address, or kernel address,
// depending on usr_src.
// Returns 0 on success, -1 on error.
int
either_copyin(void *dst, int user_src, uint64 src, uint64 len)
{
  // struct proc *p = myproc();
  if(user_src){
    // return copyin(p->pagetable, dst, src, len);
    return copyin2(dst, src, len);
  } else {
    memmove(dst, (char*)src, len);
    return 0;
  }
}

// Print a process listing to console.  For debugging.
// Runs when user types ^P on console.
// No lock to avoid wedging a stuck machine further.
void
procdump(void)
{
  static char *states[] = {
  [UNUSED]    "unused",
  [SLEEPING]  "sleep ",
  [RUNNABLE]  "runble",
  [RUNNING]   "run   ",
  [ZOMBIE]    "zombie"
  };
  struct proc *p;
  char *state;

  printf("\nPID\tSTATE\tNAME\tMEM\n");
  for(p = proc; p < &proc[NPROC]; p++){
    if(p->state == UNUSED)
      continue;
    if(p->state >= 0 && p->state < NELEM(states) && states[p->state])
      state = states[p->state];
    else
      state = "???";
    printf("%d\t%s\t%s\t%d", p->pid, state, p->name, p->sz);
    printf("\n");
  }
}

uint64
procnum(void)
{
  int num = 0;
  struct proc *p;

  for (p = proc; p < &proc[NPROC]; p++) {
    if (p->state != UNUSED) {
      num++;
    }
  }

  return num;
}


void
pqueue_init(struct queue_t *queue, int priority)
{
   queue->priority=priority;     //为方便输出debug，加入变量priority
   queue->size=0;
   for(int i=0; i<NPROC; i++)    //初始时指针为0
      queue->proc[i]=0;
}

void
pinit(void)
{
  //initlock(&ptable.lock, "ptable");
				
  pqueue_init(&pqueue_1,1);
  pqueue_init(&pqueue_2,2);
  pqueue_init(&pqueue_3,3);
  pqueue_init(&pqueue_4,4);


}

void 
enqueue(struct queue_t *q, struct proc *p)
{
    //printf("proc %d enqueueing %d.\n", p->pid, q->priority);
    int pos;
    pos = q->size - 1;
    while(pos >= 0 && q->proc[pos]->ctime > p->ctime){   //从后往前扫描
        q->proc[pos+1] = q->proc[pos];    //如果目前进程比新插入进程到达晚，就把它向后挪一个位置
		pos--;
    }
    // 目前pos指向的进程比插入的进程到达时间早或相同。
    q->proc[pos+1] = p;   
    q->size ++;   
}

void
removeq(struct queue_t* q, struct proc *p)
{
   //printf("removing %d from q: %d\n",p->pid, q->priority);
   int pos;
   int found = 0;

   pos = 0;
   while(pos < q->size){
      if(q->proc[pos]==p){   //从前往后，找到该进程
          found=1;
          break;
      }
      else{
        pos++;
      }
   }
   if(found){   //若找到，此时pos指向要删除的进程
      while(pos < q->size-1){    //从 pos 到倒数第二个进程位置
            q->proc[pos] = q->proc[pos+1];  //将所有进程往前挪一个位置
            pos++;
      }
      q->proc[q->size-1] = 0;
      q->size --;
    }
   
}

struct proc*
headq(struct queue_t* q)
{
    struct proc* p = 0;
    int pos;

    for(pos = 0; pos < q->size; pos++){
      if(q->proc[pos]->state == RUNNABLE){    //找到第一个RUNNABLE进程，注意等待队列是从到达时间小到大排序的
        p = q->proc[pos];
        break;
      }
    }

    return p;
}


void 
update()
{
  struct proc *p;

  for (p = proc; p < &proc[NPROC]; p++)
  {
    acquire(&p->lock); 
    switch (p->state)
    {              // 遍历整个表，判断进程状态
    case RUNNING:  // 运行中
      p->rutime++; // 则运行时间增加一个时间片
      break;
    case SLEEPING: // 休眠中
      p->sltime++; // 休眠时间增加一个时间片
      break;
    case RUNNABLE: // 等待中
      p->retime++; // 等待时间增加一个时间片
      break;
    default:; // default，有ZOMBIE\EMBRYO\UNUSED，这些状态我们不关心
    }
    release(&p->lock); // 释放锁
  }
  
}

int set_priority(int pid, int priority)
{
  struct proc *p = 0;
  
  for (p = proc; p < &proc[NPROC]; p++)
  {
    acquire(&p->lock);

    if (p->pid == pid)
    {

      if (p->priority != priority)
      {
        switch (p->priority)
        {
        case 1:
          removeq(&pqueue_1, p);
          break;
        case 2:
          removeq(&pqueue_2, p);
          break;
        case 3:
          removeq(&pqueue_3, p);
          break;
        case 4:
          removeq(&pqueue_4, p);
          break;
        default:;
        }
        switch (priority)
        {
        case 1:
          enqueue(&pqueue_1, p);
          break;
        case 2:
          enqueue(&pqueue_2, p);
          break;
        case 3:
          enqueue(&pqueue_3, p);
          break;
        case 4:
          enqueue(&pqueue_4, p);
          break;
        default:;
        }
      }

      p->priority = priority;

      release(&p->lock);
      return 0;
    }
    release(&p->lock);
  }

  
  return -1;
}
