#include "include/types.h"
#include "include/param.h"
#include "include/memlayout.h"
#include "include/riscv.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/syscall.h"
#include "include/sysinfo.h"
#include "include/kalloc.h"
#include "include/vm.h"
#include "include/string.h"
#include "include/printf.h"
#include "include/signal.h"

/**
 * @brief 设置信号处理动作
 * 
 * 此函数用于设置指定信号的处理动作。它允许用户为特定信号注册一个新的信号处理函数，
 * 并可选择获取当前的信号处理动作。
 * 
 * @param signum 要设置的信号编号
 * @param action 指向新的信号处理动作的指针
 * @param old_action 指向用于存储当前信号处理动作的指针
 * @return 成功时返回0，失败时返回-1
 */
int set_sigaction(int signum, Sigaction const *action, Sigaction *old_action)
{
    // 获取当前进程的指针
    struct proc * p = myproc();

    // 检查信号编号是否有效，以及输入参数是否为空
    if(signum > MAX_SIG || action == 0 || old_action == 0 || signum == SIGKILL || signum == SIGSTOP){
        // 如果信号编号无效或输入参数为空，返回错误代码-1
        return -1;
    }

    // 如果当前进程已经有一个信号处理函数，并且用户提供了old_action指针，
    // 则将当前的信号处理动作复制到old_action指向的内存中
    if(p->sigaction[signum].sig_handler != NULL && old_action != NULL){
        *old_action = p->sigaction[signum];
    }

    // 如果用户提供了新的信号处理动作指针，则将新的信号处理动作复制到当前进程的sigaction数组中
    if(action != NULL){
        p->sigaction[signum] = *action;
    }

    // 成功设置信号处理动作，返回0
    return 0;
}


/**
 * @brief 设置或获取当前进程的信号屏蔽字
 * 
 * 此函数用于设置或获取当前进程的信号屏蔽字。它允许用户阻塞或解除阻塞某些信号，
 * 并可选择获取当前的信号屏蔽字。
 * 
 * @param how 指示如何更改信号屏蔽字，取值为 SIG_BLOCK、SIG_UNBLOCK 或 SIG_SETMASK
 * @param set 指向要设置的信号屏蔽字的指针
 * @param oldset 指向用于存储当前信号屏蔽字的指针
 * @return 成功时返回0，失败时返回-1
 */
int sigprocmask(int how, sigset_t const *set, sigset_t *oldset){
    // 获取当前进程的指针
    struct proc *p = myproc();

    // 如果用户提供了oldset指针，则将当前的信号屏蔽字复制到oldset指向的内存中
    if (oldset != NULL) {
        oldset->val[0] = p->signal_mask.val[0];
    }

    // 如果用户没有提供set指针，则直接返回0，表示获取当前信号屏蔽字成功
    if(set == NULL) 
        return 0;

    // 根据how的值，更新当前进程的信号屏蔽字
    switch (how) {
        // 将set中的信号添加到当前信号屏蔽字中
        case SIG_BLOCK: 
            p->signal_mask.val[0] |= set->val[0];
            break;
        // 从当前信号屏蔽字中移除set中的信号
        case SIG_UNBLOCK: 
            p->signal_mask.val[0] &= ~(set->val[0]);
            break;
        // 将当前信号屏蔽字设置为set中的信号
        case SIG_SETMASK: 
            p->signal_mask.val[0] = set->val[0];
            break;
        // 如果how的值无效，则不做任何操作
        default: 
            break;
    }
	
    // SIGTERM SIGKILL SIGSTOP cannot be masked 
    // 确保 SIGTERM、SIGKILL 和 SIGSTOP 信号不会被屏蔽
    p->signal_mask.val[0] &= 1ul << SIGTERM | 1ul << SIGKILL | 1ul << SIGSTOP;
    return 0;
}


/**
 * @brief 处理当前进程的挂起信号
 * 
 * 此函数用于处理当前进程的挂起信号。它遍历所有可能的信号，检查是否有信号被挂起且未被屏蔽，
 * 如果有，则设置信号处理函数并准备执行。
 * 
 * @return 无返回值
 */
void handle_pending_signals(){
    // 获取当前进程的指针
    struct proc *p = myproc();

    // 遍历所有可能的信号
    for(int sig =0; sig < MAX_SIG; sig++){
        // 计算当前信号的位掩码
        uint64 sigBit = 1<<sig;

        // 检查当前信号是否被挂起且未被屏蔽
        if((p->signal.val[0] & sigBit) && !(p->signal_mask.val[0] & sigBit)){
            // 为信号处理函数分配内存
            p->sig_trapframe = kalloc();

            // 复制当前进程的trapframe到sig_trapframe
            memcpy(p->sig_trapframe, p->trapframe, sizeof(struct trapframe));

            // 设置trapframe的epc为信号处理函数的地址
            p->trapframe->epc = (uint64)p->sigaction[sigBit].sig_handler;

            // 设置trapframe的ra为信号处理函数的返回地址
            p->trapframe->ra = (uint64)SIGTRAMPOLINE;

            // 调整trapframe的sp指针，为信号处理函数的栈帧腾出空间
            p->trapframe->sp = p->trapframe->sp - PGSIZE;

            // 清除当前信号的屏蔽位
            p->signal_mask.val[0] &= ~sigBit;

            // 如果所有信号都未被屏蔽，则清除进程的killed标志
            if(p->signal_mask.val[0] == 0){
                p->killed = 0;
            }

            // 处理完一个信号后，跳出循环
            break;
        }
    }

    // 函数执行完毕，返回
    return;
}



/**
 * @brief 从信号处理程序返回
 * 
 * 此函数用于从信号处理程序返回到原始执行点。它恢复进程的trapframe，
 * 释放为信号处理程序分配的内存，并返回信号处理程序的返回值。
 * 
 * @return 返回信号处理程序的返回值
 */
uint64 rt_sigreturn(void){
    // 获取当前进程的指针
    struct proc *p = myproc();

    // 恢复进程的trapframe
    memcpy(p->trapframe, p->sig_trapframe, sizeof(struct trapframe));

    // 释放为信号处理程序分配的内存
    kfree(p->sig_trapframe);

    // 重置sig_trapframe指针
    p->sig_trapframe = 0;

    // 返回信号处理程序的返回值
    return p->trapframe->a0;
}

