//
// File-system system calls.
// Mostly argument checking, since we don't trust
// user code, and calls into file.c and fs.c.
//


#include "include/types.h"
#include "include/riscv.h"
#include "include/param.h"
#include "include/stat.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/sleeplock.h"
#include "include/file.h"
#include "include/pipe.h"
#include "include/fcntl.h"
#include "include/fat32.h"
#include "include/syscall.h"
#include "include/string.h"
#include "include/printf.h"
#include "include/vm.h"

extern struct dirent *enameparent_with_fd(int fd, char *path, char *name);

// Fetch the nth word-sized system call argument as a file descriptor
// and return both the descriptor and the corresponding struct file.
static int
argfd(int n, int *pfd, struct file **pf)
{
  int fd;
  struct file *f;

  if(argint(n, &fd) < 0)
    return -1;
  if(fd < 0 || fd >= NOFILE || (f=myproc()->ofile[fd]) == NULL)
    return -1;
  if(pfd)
    *pfd = fd;
  if(pf)
    *pf = f;
  return 0;
}

// Allocate a file descriptor for the given file.
// Takes over file reference from caller on success.
static int
fdalloc(struct file *f)
{
  int fd;
  struct proc *p = myproc();

  for(fd = 0; fd < NOFILE; fd++){
    if(p->ofile[fd] == 0){
      p->ofile[fd] = f;
      return fd;
    }
  }
  return -1;
}

//如果目标文件描述符 new 已被占用,关闭目标文件描述符 
//分配目标文件描述符 new,将 old 文件描述符复制到目标位置
static int
fdalloc_new(struct file *f, int new)
{
    struct proc *p = myproc(); // 获取当前进程结构

    // 检查文件描述符范围是否合法
    if (new < 0 || new >= NOFILE) {
        return -1; // 文件描述符超出范围
    }

    // 如果目标描述符已经被占用，先关闭它
    if (p->ofile[new] != 0) {
        fileclose(p->ofile[new]); // 关闭旧文件
        p->ofile[new] = 0;        // 清空描述符表对应的条目
    }

    // 分配新的文件描述符
    p->ofile[new] = f; // 将目标文件描述符指向文件 `f`
    return new;        // 返回目标文件描述符
}


uint64
sys_dup(void)
{
  struct file *f;
  int fd;

  if(argfd(0, 0, &f) < 0)
    return -1;
  if((fd=fdalloc(f)) < 0)
    return -1;
  filedup(f);
  return fd;
}


uint64
sys_dup2(void)
{
  struct file *f;
  int new;
  int fd;
  if(argfd(0, 0, &f) < 0 || argint(1, &new) < 0)
    return -1;
  if((fd=fdalloc_new(f, new)) < 0)
    return -1;
  filedup(f);
  return fd;
}


uint64
sys_read(void)
{
  struct file *f;
  int n;
  uint64 p;

  if(argfd(0, 0, &f) < 0 || argint(2, &n) < 0 || argaddr(1, &p) < 0)
    return -1;
  return fileread(f, p, n);
}

uint64
sys_write(void)
{
  struct file *f;
  int n;
  uint64 p;

  if(argfd(0, 0, &f) < 0 || argint(2, &n) < 0 || argaddr(1, &p) < 0)
    return -1;

  return filewrite(f, p, n);
}

uint64
sys_close(void)
{
  int fd;
  struct file *f;

  if(argfd(0, &fd, &f) < 0)
    return -1;
  myproc()->ofile[fd] = 0;
  fileclose(f);
  return 0;
}

uint64
sys_fstat(void)
{
  struct file *f;
  uint64 st; // user pointer to struct stat

  if(argfd(0, 0, &f) < 0 || argaddr(1, &st) < 0)
    return -1;
  return filestat(f, st);
}

static struct dirent*
create(char *path, short type, int mode)
{
  struct dirent *ep, *dp;
  char name[FAT32_MAX_FILENAME + 1];

  if((dp = enameparent(path, name)) == NULL)
    return NULL;

  if (type == T_DIR) {
    mode = ATTR_DIRECTORY;
  } else if (mode & O_RDONLY) {
    mode = ATTR_READ_ONLY;
  } else {
    mode = 0;  
  }

  elock(dp);
  if ((ep = ealloc(dp, name, mode)) == NULL) {
    eunlock(dp);
    eput(dp);
    return NULL;
  }
  
  if ((type == T_DIR && !(ep->attribute & ATTR_DIRECTORY)) ||
      (type == T_FILE && (ep->attribute & ATTR_DIRECTORY))) {
    eunlock(dp);
    eput(ep);
    eput(dp);
    return NULL;
  }

  eunlock(dp);
  eput(dp);

  elock(ep);
  return ep;
}


int has_permission(short file_uid, short file_gid, short file_mode, short proc_uid, short proc_gid, int omode) {
  // 如果是文件拥有者
  if (proc_uid == file_uid) {
    // 检查读权限
    if ((omode & O_RDONLY) && !(file_mode & 0400)) return 0;
    // 检查写权限
    if ((omode & O_WRONLY) && !(file_mode & 0200)) return 0;
    return 1; // 拥有者权限通过
  }

  // 如果是文件所属组
  if (proc_gid == file_gid) {
    // 检查读权限
    if ((omode & O_RDONLY) && !(file_mode & 0040)) return 0;
    // 检查写权限
    if ((omode & O_WRONLY) && !(file_mode & 0020)) return 0;
    return 1; // 组权限通过
  }

  // 如果是其他用户
  // 检查读权限
  if ((omode & O_RDONLY) && !(file_mode & 0004)) return 0;
  // 检查写权限
  if ((omode & O_WRONLY) && !(file_mode & 0002)) return 0;

  return 1; // 其他用户权限通过
}


uint64
sys_open(void)
{
  char path[FAT32_MAX_PATH];
  int fd, omode;
  struct file *f;
  struct dirent *ep;
  struct proc *p = myproc(); // 获取当前进程

  if(argstr(0, path, FAT32_MAX_PATH) < 0 || argint(1, &omode) < 0)
    return -1;

  if(omode & O_CREATE){
    ep = create(path, T_FILE, omode);
    if(ep == NULL){
      return -1;
    }
  } else {
    if((ep = ename(path)) == NULL){
      return -1;
    }
    elock(ep);
    if((ep->attribute & ATTR_DIRECTORY) && omode != O_RDONLY){
      eunlock(ep);
      eput(ep);
      return -1;
    }
  }
  // ====== 新增权限检查逻辑 ======
  // 获取文件的 UID、GID 和权限
  short file_uid = ep->uid;
  short file_gid = ep->gid;
  short file_mode = ep->mode;

  // 检查读写权限
  if (!has_permission(file_uid, file_gid, file_mode, p->uid, p->gid, omode)) {
    eunlock(ep);
    eput(ep);
    return -1; // 没有权限
  }
  // ====== 权限检查结束 ======

  // 分配文件表和文件描述符
  if((f = filealloc()) == NULL || (fd = fdalloc(f)) < 0){
    if (f) {
      fileclose(f);
    }
    eunlock(ep);
    eput(ep);
    return -1;
  }

  // 如果指定了 O_TRUNC 且不是目录，则清空文件内容
  if(!(ep->attribute & ATTR_DIRECTORY) && (omode & O_TRUNC)){
    etrunc(ep);
  }

  f->type = FD_ENTRY;
  f->off = (omode & O_APPEND) ? ep->file_size : 0;
  f->ep = ep;
  f->readable = !(omode & O_WRONLY);
  f->writable = (omode & O_WRONLY) || (omode & O_RDWR);

  // 记录文件的权限和拥有者信息
  f->uid = ep->uid;
  f->gid = ep->gid;
  f->mode = ep->mode;

  eunlock(ep);

  return fd;
}


/* 
  sys_chmod 用于修改文件的权限模式
  接受文件路径和新的权限位作为参数，并更新对应文件的 mode 字段
 */
int sys_chmod(void) {
  char path[FAT32_MAX_PATH];
  int mode;
  struct dirent *ep;

  // 从用户空间获取文件路径和权限
  if (argstr(0, path, FAT32_MAX_PATH) < 0 || argint(1, &mode) < 0)
    return -1;

  // 检查权限值是否合理
  if (mode < 0 || mode > 0777) {
    return -1; // 无效的权限值
  }


  if ((ep = ename(path)) == NULL) { // 查找文件
     
    return -1;
  }
  elock(ep);

  // 检查是否有权限修改文件（只有文件拥有者或 root 可以修改权限）
  if (myproc()->uid != ep->uid && myproc()->uid != 0) {
    eunlock(ep);
    eput(ep);
     
    return -1; // 当前用户无权限修改
  }

  // 更新文件权限
  ep->mode = (short)mode;
  eupdate(ep); // 写回更改到磁盘
  eunlock(ep);
  eput(ep);


  return 0; // 修改成功
}

/* 
  sys_chown 用于修改文件的拥有者（uid）和所属组（gid）
  接受文件路径、用户 ID 和组 ID 作为参数
 */
int sys_chown(void) {
  char path[FAT32_MAX_PATH];
  int uid, gid;
  struct dirent *ep;

  // 从用户空间获取文件路径、用户 ID 和组 ID
  if (argstr(0, path, FAT32_MAX_PATH) < 0 || argint(1, &uid) < 0 || argint(2, &gid) < 0)
    return -1;

  // 检查用户 ID 和组 ID 是否合理
  if (uid < 0 || uid > 65535 || gid < 0 || gid > 65535) {
    return -1; // 无效的 uid 或 gid
  }

  if ((ep = ename(path)) == NULL) { // 查找文件
     
    return -1;
  }
  elock(ep);

  // 检查是否有权限修改文件（只有文件拥有者或 root 可以修改拥有者和组）
  if (myproc()->uid != ep->uid && myproc()->uid != 0) {
    eunlock(ep);
    eput(ep);
     
    return -1; // 当前用户无权限修改
  }

  // 更新文件拥有者和所属组
  ep->uid = (short)uid;
  ep->gid = (short)gid;
  eupdate(ep); // 写回更改到磁盘
  eunlock(ep);
  eput(ep);


  return 0; // 修改成功
}



/* 
  openat的参数:
  flags表示打开文件的方式，与open中的omode一致  文件打开操作的模式标志
  mode则表示文件的属性,新建文件时需指定         文件/目录的属性标志
 */
uint64
sys_openat(void)
{
  char path[FAT32_MAX_PATH];  // 存储路径参数
  int fd,flags,mode;
  
  // 参数解析：
  // 参数 0: 文件描述符
  // 参数 1: 文件路径
  // 参数 2: 打开标志
  // 参数 3: 文件属性
  if(argint(0, &fd)<0 || argstr(1, path, FAT32_MAX_PATH) < 0 || argint(2, &flags) < 0 || argint(3, &mode) ){
    return -1;
  }

  struct dirent* entry;
  // 根据 fd 和 path 查找文件对应的目录项
  if((entry = ename_with_fd(fd, path)) == NULL){  //如果找到文件对应的目录项
      if(flags & O_CREATE){       // 如果指定了 O_CREATE，则尝试创建新文件
        if((entry = create_with_fd(fd, path, T_FILE, mode)) == NULL){
          return -1;
        }
      }
      else{
        return -1;\
      }
    }
  
  else{
    // 找到了对应的目录项
    elock(entry);
    if (!(entry->attribute & ATTR_DIRECTORY) && (flags & O_DIRECTORY)) {
        // 如果目标不是目录，但指定了 O_DIRECTORY，返回错误
        eunlock(entry);
        eput(entry);
        return -1;
    }
    if ((entry->attribute & ATTR_DIRECTORY) && (flags & O_RDONLY)) { 
        // 如果目标是目录，但未指定只读模式，返回错误
        eunlock(entry);
        eput(entry);
        return -1;
    }
  }
  // 分配文件结构
  struct file* file;
  if ((file = filealloc()) == NULL || (fd = fdalloc(file)) < 0) {
      if (file) {
          fileclose(file);
      }
      eunlock(entry);
      eput(entry);
      return -1;
  }

  // 如果目标不是目录，且指定了 O_TRUNC，则清空文件内容
  if (!(entry->attribute & ATTR_DIRECTORY) && (flags & O_TRUNC)) {
      etrunc(entry);
  }

  // 初始化文件结构
  file->type = FD_ENTRY;
  file->off = (flags & O_APPEND) ? entry->file_size : 0;
  file->ep = entry;
  file->readable = !(flags & O_WRONLY);
  file->writable = (flags & O_WRONLY) || (flags & O_RDWR);

  eunlock(entry);

  return fd;

}


uint64
sys_touch(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;

  if(argstr(0, path, FAT32_MAX_PATH) < 0 || (ep = create(path, T_FILE, 0)) == 0){
    return -1;
  }
  eunlock(ep);
  eput(ep);
  return 0;
}

uint64
sys_chdir(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;
  struct proc *p = myproc();
  
  if(argstr(0, path, FAT32_MAX_PATH) < 0 || (ep = ename(path)) == NULL){
    return -1;
  }
  elock(ep);
  if(!(ep->attribute & ATTR_DIRECTORY)){
    eunlock(ep);
    eput(ep);
    return -1;
  }
  eunlock(ep);
  eput(p->cwd);
  p->cwd = ep;
  return 0;
}

uint64
sys_pipe(void)
{
  uint64 fdarray; // user pointer to array of two integers
  struct file *rf, *wf;
  int fd0, fd1;
  struct proc *p = myproc();

  if(argaddr(0, &fdarray) < 0)
    return -1;
  if(pipealloc(&rf, &wf) < 0)
    return -1;
  fd0 = -1;
  if((fd0 = fdalloc(rf)) < 0 || (fd1 = fdalloc(wf)) < 0){
    if(fd0 >= 0)
      p->ofile[fd0] = 0;
    fileclose(rf);
    fileclose(wf);
    return -1;
  }
  // if(copyout(p->pagetable, fdarray, (char*)&fd0, sizeof(fd0)) < 0 ||
  //    copyout(p->pagetable, fdarray+sizeof(fd0), (char *)&fd1, sizeof(fd1)) < 0){
  if(copyout2(fdarray, (char*)&fd0, sizeof(fd0)) < 0 ||
     copyout2(fdarray+sizeof(fd0), (char *)&fd1, sizeof(fd1)) < 0){
    p->ofile[fd0] = 0;
    p->ofile[fd1] = 0;
    fileclose(rf);
    fileclose(wf);
    return -1;
  }
  return 0;
}

// To open console device.
uint64
sys_dev(void)
{
  int fd, omode;
  int major, minor;
  struct file *f;

  if(argint(0, &omode) < 0 || argint(1, &major) < 0 || argint(2, &minor) < 0){
    return -1;
  }

  if(omode & O_CREATE){
    panic("dev file on FAT");
  }

  if(major < 0 || major >= NDEV)
    return -1;

  if((f = filealloc()) == NULL || (fd = fdalloc(f)) < 0){
    if(f)
      fileclose(f);
    return -1;
  }

  f->type = FD_DEVICE;
  f->off = 0;
  f->ep = 0;
  f->major = major;
  f->readable = !(omode & O_WRONLY);
  f->writable = (omode & O_WRONLY) || (omode & O_RDWR);

  return fd;
}

// To support ls command
uint64
sys_readdir(void)
{
  struct file *f;
  uint64 p;

  if(argfd(0, 0, &f) < 0 || argaddr(1, &p) < 0)
    return -1;
  return dirnext(f, p);
}

// get absolute cwd string
uint64
sys_getcwd(void)
{
  uint64 addr;
  if (argaddr(0, &addr) < 0)
    return -1;

  struct dirent *de = myproc()->cwd;
  char path[FAT32_MAX_PATH];
  char *s;
  int len;

  if (de->parent == NULL) {
    s = "/";
  } else {
    s = path + FAT32_MAX_PATH - 1;
    *s = '\0';
    while (de->parent) {
      len = strlen(de->filename);
      s -= len;
      if (s <= path)          // can't reach root "/"
        return -1;
      strncpy(s, de->filename, len);
      *--s = '/';
      de = de->parent;
    }
  }

  // if (copyout(myproc()->pagetable, addr, s, strlen(s) + 1) < 0)
  if (copyout2(addr, s, strlen(s) + 1) < 0)
    return -1;
  
  return 0;

}


// get absolute cwd string
uint64
sys_pwd(void)
{
  uint64 addr;
  if (argaddr(0, &addr) < 0)
    return -1;

  struct dirent *de = myproc()->cwd;
  char path[FAT32_MAX_PATH];
  char *s;
  int len;

  if (de->parent == NULL) {
    s = "/";
  } else {
    s = path + FAT32_MAX_PATH - 1;
    *s = '\0';
    while (de->parent) {
      len = strlen(de->filename);
      s -= len;
      if (s <= path)          // can't reach root "/"
        return -1;
      strncpy(s, de->filename, len);
      *--s = '/';
      de = de->parent;
    }
  }

  // if (copyout(myproc()->pagetable, addr, s, strlen(s) + 1) < 0)
  if (copyout2(addr, s, strlen(s) + 1) < 0)
    return -1;
  
  return 0;

}


// Is the directory dp empty except for "." and ".." ?
static int
isdirempty(struct dirent *dp)
{
  struct dirent ep;
  int count;
  int ret;
  ep.valid = 0;
  ret = enext(dp, &ep, 2 * 32, &count);   // skip the "." and ".."
  return ret == -1;
}

uint64
sys_remove(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;
  int len;
  if((len = argstr(0, path, FAT32_MAX_PATH)) <= 0)
    return -1;

  char *s = path + len - 1;
  while (s >= path && *s == '/') {
    s--;
  }
  if (s >= path && *s == '.' && (s == path || *--s == '/')) {
    return -1;
  }
  
  if((ep = ename(path)) == NULL){
    return -1;
  }
  elock(ep);
  if((ep->attribute & ATTR_DIRECTORY) && !isdirempty(ep)){
      eunlock(ep);
      eput(ep);
      return -1;
  }
  elock(ep->parent);      // Will this lead to deadlock?
  eremove(ep);
  eunlock(ep->parent);
  eunlock(ep);
  eput(ep);

  return 0;
}

// Must hold too many locks at a time! It's possible to raise a deadlock.
// Because this op takes some steps, we can't promise
uint64
sys_rename(void)
{
  char old[FAT32_MAX_PATH], new[FAT32_MAX_PATH];
  if (argstr(0, old, FAT32_MAX_PATH) < 0 || argstr(1, new, FAT32_MAX_PATH) < 0) {
      return -1;
  }

  struct dirent *src = NULL, *dst = NULL, *pdst = NULL;
  int srclock = 0;
  char *name;
  if ((src = ename(old)) == NULL || (pdst = enameparent(new, old)) == NULL
      || (name = formatname(old)) == NULL) {
    goto fail;          // src doesn't exist || dst parent doesn't exist || illegal new name
  }
  for (struct dirent *ep = pdst; ep != NULL; ep = ep->parent) {
    if (ep == src) {    // In what universe can we move a directory into its child?
      goto fail;
    }
  }

  uint off;
  elock(src);     // must hold child's lock before acquiring parent's, because we do so in other similar cases
  srclock = 1;
  elock(pdst);
  dst = dirlookup(pdst, name, &off);
  if (dst != NULL) {
    eunlock(pdst);
    if (src == dst) {
      goto fail;
    } else if (src->attribute & dst->attribute & ATTR_DIRECTORY) {
      elock(dst);
      if (!isdirempty(dst)) {    // it's ok to overwrite an empty dir
        eunlock(dst);
        goto fail;
      }
      elock(pdst);
    } else {                    // src is not a dir || dst exists and is not an dir
      goto fail;
    }
  }

  if (dst) {
    eremove(dst);
    eunlock(dst);
  }
  memmove(src->filename, name, FAT32_MAX_FILENAME);
  emake(pdst, src, off);
  if (src->parent != pdst) {
    eunlock(pdst);
    elock(src->parent);
  }
  eremove(src);
  eunlock(src->parent);
  struct dirent *psrc = src->parent;  // src must not be root, or it won't pass the for-loop test
  src->parent = edup(pdst);
  src->off = off;
  src->valid = 1;
  eunlock(src);

  eput(psrc);
  if (dst) {
    eput(dst);
  }
  eput(pdst);
  eput(src);

  return 0;

fail:
  if (srclock)
    eunlock(src);
  if (dst)
    eput(dst);
  if (pdst)
    eput(pdst);
  if (src)
    eput(src);
  return -1;
}

// sys_getdents64: 实现 getdents64 系统调用，用于从目录文件描述符中读取目录项
// 并将它们返回到用户提供的缓冲区中。
// 返回成功读取的字节数，失败时返回 -1。

uint64 sys_getdents64(void)
{
    struct proc *p = myproc();  // 获取当前进程
    struct file *f;             // 文件指针，用于操作目录
    int fd;                     // 文件描述符
    uint64 buf, len;            // 用户提供的缓冲区地址和缓冲区长度

    // 解析用户传入的参数：
    // 参数 0: 文件描述符
    // 参数 1: 用户缓冲区地址
    // 参数 2: 用户缓冲区大小
    if (argfd(0, &fd, &f) < 0 || argaddr(1, &buf) < 0 || argaddr(2, &len) < 0) {
        return -1;  // 如果参数解析失败，则返回 -1
    }

    // 检查文件是否可读，以及是否是目录
    if (f->readable == 0 || !(f->ep->attribute & ATTR_DIRECTORY)) {
        return -1;  // 如果不是可读的目录文件，则返回 -1
    }

    int err;                       // 用于存储读取目录项时的错误状态
    struct dirent dste;            // 文件系统内部的目录项结构
    dirent64 dir64;                // 用于返回给用户的 linux_dirent64 结构
    int copysize = 0;              // 已经复制到用户缓冲区的字节数
    int count = 0;                 // 当前目录项占用的条目数量

    elock(f->ep);                  // 加锁，确保读取目录项时的一致性

    // 遍历目录内容
    while (1) {
        dir64.d_off = f->off;      // 设置当前目录项的偏移量
        // 跳过空目录项
        while ((err = enext(f->ep, &dste, f->off, &count)) == 0) {
            f->off += count * 32;  // 如果是空目录项，跳过，更新文件偏移量
        }

        if (err == -1) {           // 如果到达目录末尾，退出循环
            break;
        }

        f->off += count * 32;      // 读取非空目录项后，更新文件偏移量

        // 填充 linux_dirent64 结构
        dir64.d_type = (dste.attribute & ATTR_DIRECTORY) ? T_DIR : T_FILE;  // 判断目录项类型
        dir64.d_ino = 0;           // 暂时未用，设置为 0
        strncpy(dir64.d_name, dste.filename, sizeof(dste.filename));  // 拷贝文件名
        dir64.d_reclen = strlen(dir64.d_name) + sizeof(dirent64) - sizeof(dir64.d_name) + 1;  // 计算记录长度

        // 检查缓冲区是否足够存储当前目录项
        if (dir64.d_reclen > len) {
            break;  // 如果缓冲区不足，停止读取
        }

        int copySize = dir64.d_reclen;  // 记录当前目录项的大小

        // 将目录项从内核空间复制到用户空间
        if (copyout(p->pagetable, buf, (char *)&dir64, copySize) < 0) {
            eunlock(f->ep);  // 如果复制失败，解锁并返回错误
            return -1;
        }

        buf += copySize;       // 更新用户缓冲区指针
        len -= copySize;       // 更新剩余缓冲区大小
        copysize += copySize;  // 累计已复制的总字节数
    }

    eunlock(f->ep);  // 解锁目录项
    return copysize; // 返回成功写入用户缓冲区的字节数
}

// unlinkat用于删除文件或空目录
// 返回值：成功时返回 0，失败时返回 -1

uint64 
sys_unlinkat(void)
{
    char path[FAT32_MAX_PATH];      // 用于存储路径参数
    struct dirent *ep;              // 指向目标文件的目录项
    char name[FAT32_MAX_FILENAME];  // 保存路径中的文件名
    struct dirent *parent;          // 指向父目录的目录项
    int dirfd;                      // 文件描述符

    // 解析用户参数
    if (argint(0, &dirfd) < 0 || argstr(1, path, FAT32_MAX_PATH) < 0) {
        return -1;  // 参数解析失败
    }

    // 解析路径并获取父目录和文件名
    if ((parent = enameparent_with_fd(dirfd, path, name)) == NULL) {
        return -1;  // 父目录解析失败，返回 -1
    }

    // 加锁父目录
    elock(parent);

    // 查找目标目录项
    if ((ep = dirlookup(parent, name, 0)) == NULL) {
        eunlock(parent);
        eput(parent);
        return -1;  // 文件不存在，返回 -1
    }

    // 锁定目标文件
    elock(ep);

    // 检查目标是否是目录
    if (ep->attribute & ATTR_DIRECTORY) {
        // 如果是目录，返回错误
        eunlock(ep);
        eput(ep);
        eunlock(parent);
        eput(parent);
        return -1;  // 目标是目录，返回 -1
    }

    // 删除目录项
    eremove(ep);

    // 解锁并释放引用
    eunlock(ep);
    eput(ep);
    eunlock(parent);
    eput(parent);

    return 0;  // 删除成功，返回 0
}




uint64
sys_mkdir(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;

  if(argstr(0, path, FAT32_MAX_PATH) < 0 || (ep = create(path, T_DIR, 0)) == 0){
    return -1;
  }
  eunlock(ep);
  eput(ep);
  return 0;
}

// mkdirat用于在指定路径创建目录
// 返回值：成功时返回 0，失败时返回 -1

uint64 
sys_mkdirat(void)
{
    char path[FAT32_MAX_PATH];      // 用于存储路径参数
    struct dirent *parent;          // 父目录的目录项
    struct dirent *ep;              // 指向新建的目录项
    char name[FAT32_MAX_FILENAME];  // 保存路径中的目录名
    int dirfd, mode;                // 文件描述符和模式

    // 参数解析
    if (argint(0, &dirfd) < 0 || argstr(1, path, FAT32_MAX_PATH) < 0 || argint(2, &mode) < 0) {
        return -1;  // 参数解析失败，返回 -1
    }

    // 解析路径并获取父目录和目录名
    if ((parent = enameparent_with_fd(dirfd, path, name)) == NULL) {
        return -1;  // 父目录解析失败，返回 -1
    }

    // 加锁父目录
    elock(parent);

    // 检查是否同名目录或文件已存在
    if ((ep = dirlookup(parent, name, 0)) != NULL) {
        eunlock(parent);
        eput(parent);
        eput(ep);
        return -1;  // 如果同名目录或文件已存在，返回 -1
    }

    // 创建新目录
    if ((ep = ealloc(parent, name, ATTR_DIRECTORY)) == NULL) {
        eunlock(parent);
        eput(parent);
        return -1;  // 如果分配失败，返回 -1
    }

    // 设置新目录的模式（目前仅存储模式值，不具体处理权限）
    ep->mode = mode;

    // 解锁父目录并释放其引用计数
    eunlock(parent);
    eput(parent);

    // 解锁新建目录项并释放其引用计数
    eunlock(ep);
    eput(ep);

    return 0;  // 成功返回 0
}


// by chong
uint64
sys_mmap(void)
{
  uint64 start;
  uint64 len;
  int prot;
  int flags;
  int fd;
  int off;
  // 从用户接收地址，验证是否有效后获取数据
  if(argaddr(0, &start) < 0)
    return -1;
  if(argaddr(1, &len) < 0)
    return -1;
  if(argint(2, &prot) < 0)
    return -1;
  if(argint(3, &flags) < 0)
    return -1;
  if(argfd(4, &fd, NULL) < 0)
    return -1;
  if(argint(5, &off) < 0)
    return -1;
  
  return do_mmap(start, len, prot, flags, fd, off);
}

// by chong
uint64
sys_munmap(void)
{
  uint64 start;
  uint64 len;
  // 获取虚拟地址起始位置和需要解除映射的大小
  if(argaddr(0, &start) < 0 || argaddr(1, &len) < 0){
    return -1;
  }
  return do_munmap(start, len);
}