
#include "include/types.h"
#include "include/riscv.h"
#include "include/param.h"
#include "include/memlayout.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/defs.h"
#include "include/syscall.h"
#include "include/timer.h"
#include "include/kalloc.h"
#include "include/string.h"
#include "include/printf.h"
#include "include/date.h"
#include "include/utsname.h"
#include "include/signal.h"

extern int exec(char *path, char **argv);

uint64
sys_exec(void)
{
  char path[FAT32_MAX_PATH], *argv[MAXARG];
  int i;
  uint64 uargv, uarg;

  if(argstr(0, path, FAT32_MAX_PATH) < 0 || argaddr(1, &uargv) < 0){
    return -1;
  }
  memset(argv, 0, sizeof(argv));
  for(i=0;; i++){
    if(i >= NELEM(argv)){
      goto bad;
    }
    if(fetchaddr(uargv+sizeof(uint64)*i, (uint64*)&uarg) < 0){
      goto bad;
    }
    if(uarg == 0){
      argv[i] = 0;
      break;
    }
    argv[i] = kalloc();
    if(argv[i] == 0)
      goto bad;
    if(fetchstr(uarg, argv[i], PGSIZE) < 0)
      goto bad;
  }

  int ret = exec(path, argv);

  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);

  return ret;

 bad:
  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);
  return -1;
}




uint64
sys_exit(void)
{
  int n;
  if(argint(0, &n) < 0)
    return -1;
  exit(n);
  return 0;  // not reached
}

uint64
sys_getpid(void)
{
  return myproc()->pid;
}

uint64
sys_fork(void)
{
  return fork();
}

uint64
sys_wait(void)
{
  uint64 p;
  if(argaddr(0, &p) < 0)
    return -1;
  return wait(p);
}

uint64
sys_waitpid(void)
{
  int pid,options;
  uint64 status;
  if(argint(0, &pid) < 0)
    return -1;
  if(argaddr(1, &status) < 0)
    return -1;
  if(argint(2,&options) < 0)
    return -1;
  return waitpid(pid,status,options);
}

uint64
sys_sbrk(void)
{
  int addr;
  int n;

  if(argint(0, &n) < 0)
    return -1;
  addr = myproc()->sz;
  if(growproc(n) < 0)
    return -1;
  return addr;
}

uint64
sys_sleep(void)
{
  int n;
  uint ticks0;

  if(argint(0, &n) < 0)
    return -1;
  acquire(&tickslock);
  ticks0 = ticks;
  while(ticks - ticks0 < n){
    if(myproc()->killed){
      release(&tickslock);
      return -1;
    }
    sleep(&ticks, &tickslock);
  }
  release(&tickslock);
  return 0;
}

uint64
sys_kill(void)
{
  int pid;

  if(argint(0, &pid) < 0)
    return -1;
  return kill(pid);
}

// return how many clock tick interrupts have occurred
// since start.
uint64
sys_uptime(void)
{
  uint xticks;

  acquire(&tickslock);
  xticks = ticks;
  release(&tickslock);
  return xticks;
}

uint64
sys_trace(void)
{
  int mask;
  if(argint(0, &mask) < 0) {
    return -1;
  }
  myproc()->tmask = mask;
  return 0;
}

uint64
sys_date(void)
{
  struct rtcdate r;
  struct proc *curproc = myproc(); // 获取当前进程
  char *user_dst;             // 用户传递的目标地址
  uint64 time_in_seconds = *(volatile uint64 *)CLINT_MTIME / TICKS_PER_SECOND;
  convert_to_date(time_in_seconds, &r);
  if(argptr(0,(char **)&user_dst,sizeof(struct rtcdate))<0)
    return -1;

  // 从 CLINT 获取当前时间（假设有 get_time 函数实现）
    uint64 ticks = *(volatile uint64 *)CLINT_MTIME;  // 读取 CLINT 的 mtime
    uint64 seconds = ticks / TICKS_PER_SECOND;       // 转换为秒
    convert_to_date(seconds, &r);                    // 将秒转换为日期格式

    // 将转换后的日期写入用户空间
    if (copyout(curproc->pagetable, (uint64)user_dst, (char *)&r, sizeof(r)) < 0) {
        return -1;  // 如果写入用户空间失败，返回错误
    }
  // cmostimer(r);
  return 0;
}

uint64 sys_uname(void) {
    struct utsname *uname_buf;
    // 从用户空间获取指针
    if (argaddr(0, (uint64*)&uname_buf) < 0)
        return -1;


    // 填充 uname 信息
    struct utsname info;
    safestrcpy(info.sysname, "xv6", sizeof(info.sysname));
    safestrcpy(info.nodename, "defult", sizeof(info.nodename));
    safestrcpy(info.release, "1.0", sizeof(info.release));
    safestrcpy(info.version, "xv6-riscv-k210", sizeof(info.version));
    safestrcpy(info.machine, UTS_MACHINE, sizeof(info.machine));

    // 将数据从内核空间拷贝到用户空间
    if (copyout(myproc()->pagetable, (uint64)uname_buf, (char *)&info, sizeof(info)) < 0)
        return -1;

    return 0;
}

int
sys_waitsch(void)
{
    int *rutime;
    int *retime; 
    int *sltime;
    int pid;
    rutime=retime=sltime=0;
    if(argptr(0, (char**)&rutime, sizeof(int)) < 0)
        return -1;
    if(argptr(1, (char**)&retime, sizeof(int)) < 0)
        return -1;
    if(argptr(2, (char**)&sltime, sizeof(int)) < 0)
        return -1;
    pid = waitSH(rutime, retime, sltime);
    if(pid!=-1)    //如果有子进程退出，打印出它的pid，运行时间、等待时间和休眠时间。
        printf("pid: %d, runtime:%d, ready time: %d, sleep time %d\n", pid, *rutime, *retime, *sltime);
    return pid;
} 


int
sys_set_priority(void)
{
    int pid;
    int priority;
    if(argint(0, &pid) < 0)
        return -1;
    if(argint(1, &priority) < 0)
        return -1;
    return set_priority(pid, priority);
}


uint64
sys_brk(void)
{
  int addr;
  int n;

  if(argint(0, &n) < 0)
    return -1;
  addr = myproc()->sz;
  if (n == 0)
  {
    return addr;
  }
  if (n >= addr)
  {
    if(growproc(n - addr) < 0)
      return -1;
    else return 0;
  }
  return -1;
}


uint64
sys_clone(void)
{
  uint64 func;  // 线程入口函数地址
  uint64 stack; // 用户栈地址
  uint64 arg;
  if (argaddr(0, &stack) < 0 ||argaddr(1, &func) < 0 || argaddr(2,&arg))
  {
    return -1;
  }

  // 调用 proc.c 中的 clone 函数
  return clone(stack, func, arg);
}

uint64
sys_thread_exit(void)
{
  thread_exit();

  return 0;
}

uint64
sys_thread_join(void)
{
  int pid;
  if (argint(0, &pid) < 0)
    return -1;
  return thread_join(pid);
}

uint64
sys_getppid(void)
{
  struct proc *p = myproc();
  acquire(&p->lock);
  uint64 ppid = p->parent->pid;
  release(&p->lock);
  return ppid;
}

uint64
sys_execve(void)
{
    char path[FAT32_MAX_PATH], *argv[MAXARG], *envp[MAXARG];
    uint64 uargv, uenvp, uarg, uenv;
    int i;

    // 1. 获取用户传入的路径、参数地址 (argv) 和环境变量地址 (envp)
    if (argstr(0, path, FAT32_MAX_PATH) < 0 || 
        argaddr(1, &uargv) < 0 || 
        argaddr(2, &uenvp) < 0) {
        return -1;
    }

    // 初始化 argv 和 envp
    memset(argv, 0, sizeof(argv));
    memset(envp, 0, sizeof(envp));

    // 2. 解析 argv 参数列表
    for (i = 0; i < MAXARG; i++) {
        if (fetchaddr(uargv + sizeof(uint64) * i, &uarg) < 0) { // 从用户空间读取指针
            goto bad;
        }

        if (uarg == 0) { // 遇到 NULL 表示结束
            argv[i] = 0;
            break;
        }

        argv[i] = kalloc(); // 分配内核缓冲区
        if (argv[i] == 0) {
            goto bad;
        }

        if (fetchstr(uarg, argv[i], PGSIZE) < 0) { // 从用户空间拷贝字符串到内核
            kfree(argv[i]); // 释放已分配的缓冲区
            argv[i] = 0;
            goto bad;
        }
    }

    if (i == MAXARG) { // 检查是否超出 MAXARG
        goto bad;
    }

    // 3. 解析 envp 环境变量列表
    for (i = 0; i < MAXARG; i++) {
        if (fetchaddr(uenvp + sizeof(uint64) * i, &uenv) < 0) { // 从用户空间读取指针
            goto bad;
        }

        if (uenv == 0) { // 遇到 NULL 表示结束
            envp[i] = 0;
            break;
        }

        envp[i] = kalloc(); // 分配内核缓冲区
        if (envp[i] == 0) {
            goto bad;
        }

        if (fetchstr(uenv, envp[i], PGSIZE) < 0) { // 从用户空间拷贝字符串到内核
            kfree(envp[i]); // 释放已分配的缓冲区
            envp[i] = 0;
            goto bad;
        }
    }

    if (i == MAXARG) { // 检查是否超出 MAXARG
        goto bad;
    }

    // 4. 调用核心 exec 函数
    int ret = exec(path, argv); // 核心函数暂未支持 envp，需扩展

    // 5. 释放 argv 和 envp 的内存
    for (i = 0; i < MAXARG && argv[i] != 0; i++) {
        kfree(argv[i]);
    }
    for (i = 0; i < MAXARG && envp[i] != 0; i++) {
        kfree(envp[i]);
    }

    return ret;

bad:
    // 释放已分配的内存
    for (i = 0; i < MAXARG && argv[i] != 0; i++) {
        kfree(argv[i]);
    }
    for (i = 0; i < MAXARG && envp[i] != 0; i++) {
        kfree(envp[i]);
    }
    return -1;
}


uint64
sys_rt_sigprocmask() {
 int how;
	uint64 uptr_set, uptr_oldset;

	sigset_t set, oldset;

	argint(0, &how);
	argaddr(1, &uptr_set);
	argaddr(2, &uptr_oldset);

	if (uptr_set && copyin(myproc()->pagetable, (char*)&set, uptr_set, sizeof(sigset_t)) < 0) {
		return -1;
	}

	if (sigprocmask(how, &set, uptr_oldset ? &oldset : NULL)) {
		return -1;
	}

	if (uptr_oldset && copyout(myproc()->pagetable, uptr_oldset, (char*)&oldset, sizeof(sigset_t)) < 0) {
		return -1;
	}


  return 0;
}
uint64
sys_rt_sigaction() {
 int signum;
	Sigaction* ptr_act = NULL;		// struct sigaction const *act
	Sigaction* ptr_oldact = NULL;		// struct sigaction *oldact

	argint(0, &signum);
  if (signum < 0) {
      return -1;
  }
	argaddr(1, (uint64*)&ptr_act);
	argaddr(2, (uint64*)&ptr_oldact);

	// copy struct sigaction from user space 
	Sigaction act = {0};
	Sigaction oldact = {0};

	if (ptr_act){
		if(copyin(myproc()->pagetable, (char*)&(act), (uint64)ptr_act, sizeof(Sigaction)) < 0 ) {
			return -1;
		}
	}

	if (set_sigaction(signum, ptr_act ? &act : NULL, ptr_oldact ? &oldact : NULL) < 0) {
		return -1;
	}

	if (ptr_oldact) {
        if (copyout(myproc()->pagetable, (uint64)ptr_oldact, (char*)&(oldact), sizeof(Sigaction)) < 0 ){
			return -1;
		}
	}
	// debug_print("sys_rt_sigaction: signum = %d, act fp:%p \n", signum, act.__sigaction_handler.sa_handler);
	return 0;
}
uint64
sys_rt_sigreturn(void){
	return rt_sigreturn();
}
